
from typing import TYPE_CHECKING

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QFrame

from .base_elements import PortTypesViewFlag

from .ui.type_filter_frame import Ui_FrameTypesFilter

if TYPE_CHECKING:
    from patchbay_manager import PatchbayManager


class TypeFilterFrame(QFrame):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.ui = Ui_FrameTypesFilter()
        self.ui.setupUi(self)
        
        self.ui.checkBoxAudioFilter.really_clicked.connect(
            self._check_box_audio_right_clicked)
        self.ui.checkBoxMidiFilter.really_clicked.connect(
            self._check_box_midi_right_clicked)
        self.ui.checkBoxCvFilter.really_clicked.connect(
            self._check_box_cv_right_clicked)
        self.ui.checkBoxAlsaFilter.really_clicked.connect(
            self._check_box_alsa_right_clicked)
        
        self._patchbay_mng = None
        self._alsa_midi_enabled = False
    
    def set_patchbay_manager(self, manager: 'PatchbayManager'):
        self._patchbay_mng = manager
        self._patchbay_mng.sg.port_types_view_changed.connect(
            self._port_types_view_changed)
        self._port_types_view_changed(self._patchbay_mng.port_types_view)
    
    def enable_alsa_midi(self, yesno: bool):
        self._alsa_midi_enabled = yesno
        self.ui.checkBoxAlsaFilter.setVisible(yesno)
    
    def _change_port_types_view(self):
        if self._patchbay_mng is None:
            return
        
        audio_checked = self.ui.checkBoxAudioFilter.isChecked()
        midi_checked = self.ui.checkBoxMidiFilter.isChecked()
        cv_checked = self.ui.checkBoxCvFilter.isChecked()
        alsa_checked = self.ui.checkBoxAlsaFilter.isChecked()
        
        port_types_view = PortTypesViewFlag.NONE

        if audio_checked and midi_checked and cv_checked and alsa_checked:
            port_types_view = PortTypesViewFlag.ALL
        else:
            if audio_checked:
                port_types_view |= PortTypesViewFlag.AUDIO
            if midi_checked:
                port_types_view |= PortTypesViewFlag.MIDI
            if cv_checked:
                port_types_view |= PortTypesViewFlag.CV
            if alsa_checked:
                port_types_view |= PortTypesViewFlag.ALSA
                
        if port_types_view is PortTypesViewFlag.NONE:
            if self._patchbay_mng.port_types_view is PortTypesViewFlag.AUDIO:
                port_types_view = (
                    PortTypesViewFlag.MIDI | PortTypesViewFlag.CV | PortTypesViewFlag.ALSA)
            elif self._patchbay_mng.port_types_view is PortTypesViewFlag.MIDI:
                port_types_view = (
                    PortTypesViewFlag.AUDIO | PortTypesViewFlag.CV | PortTypesViewFlag.ALSA)
            elif self._patchbay_mng.port_types_view is PortTypesViewFlag.CV:
                port_types_view = (
                    PortTypesViewFlag.AUDIO | PortTypesViewFlag.MIDI | PortTypesViewFlag.ALSA) 
            elif self._patchbay_mng.port_types_view is PortTypesViewFlag.ALSA:
                port_types_view = (
                    PortTypesViewFlag.AUDIO | PortTypesViewFlag.MIDI | PortTypesViewFlag.CV)
        
        self._patchbay_mng.change_port_types_view(port_types_view)
    
    def _exclusive_choice(self, view_flag: PortTypesViewFlag):
        if self._patchbay_mng is None:
            return
        
        if self._patchbay_mng.port_types_view is view_flag:
            self._patchbay_mng.change_port_types_view(PortTypesViewFlag.ALL)
        else:
            self._patchbay_mng.change_port_types_view(view_flag)
    
    def _check_box_audio_right_clicked(self, alternate: bool):
        if alternate:
            self._exclusive_choice(PortTypesViewFlag.AUDIO)
            return
        
        self.ui.checkBoxAudioFilter.setChecked(
            not self.ui.checkBoxAudioFilter.isChecked())
        self._change_port_types_view()
    
    def _check_box_midi_right_clicked(self, alternate: bool):
        if alternate:
            self._exclusive_choice(PortTypesViewFlag.MIDI)
            return
        
        self.ui.checkBoxMidiFilter.setChecked(
            not self.ui.checkBoxMidiFilter.isChecked())
        self._change_port_types_view()
    
    def _check_box_cv_right_clicked(self, alternate: bool):
        if alternate:
            self._exclusive_choice(PortTypesViewFlag.CV)
            return
        
        self.ui.checkBoxCvFilter.setChecked(
            not self.ui.checkBoxCvFilter.isChecked())
        self._change_port_types_view()
    
    def _check_box_alsa_right_clicked(self, alternate: bool):
        if alternate:
            self._exclusive_choice(PortTypesViewFlag.ALSA)
            return
        
        self.ui.checkBoxAlsaFilter.setChecked(
            not self.ui.checkBoxAlsaFilter.isChecked())
        self._change_port_types_view()
    
    def _port_types_view_changed(self, port_types_view: int):
        self.ui.checkBoxAlsaFilter.setVisible(
            self._patchbay_mng.alsa_midi_enabled)

        self.ui.checkBoxAudioFilter.setChecked(
            bool(port_types_view & PortTypesViewFlag.AUDIO))
        self.ui.checkBoxMidiFilter.setChecked(
            bool(port_types_view & PortTypesViewFlag.MIDI))
        self.ui.checkBoxCvFilter.setChecked(
            bool(port_types_view & PortTypesViewFlag.CV))
        self.ui.checkBoxAlsaFilter.setChecked(
            bool(port_types_view & PortTypesViewFlag.ALSA))
        


#!/usr/bin/python3 -u

# ray-jackpatch is an executable launchable in NSM (or Ray) session.
# It restores JACK connections.
# To avoid many problems, the connection processus is slow.
# Connections are made one by one, waiting to receive from jack
# the callback that a connection has been made to make the 
# following one.
# It also disconnects connections undesired in the session.
# To determinate that a connection is undesired, all the present ports
# are saved in the save file. If a connection is not saved in the file,
# and its ports were present at save time, this connection will be disconnected
# at session open.
# This disconnect behavior is (probably) not suitable if we start the 
# ray-jackpatch client once the session is already loaded.
# The only way we've got to know that the entire session is opening, 
# is to check if session_is_loaded message is received.

import os
import signal
import sys
try:
    import liblo
except ImportError:
    import pyliblo3 as liblo
import logging
import xml.etree.ElementTree as ET

from nsm_client import NsmServer, NsmCallback, Err
from bases import (EventHandler, MonitorStates, PortMode, PortType,
                   Event, JackPort, Timer, Glob, debug_conn_str)
from jack_renaming_tools import (
    port_belongs_to_client, port_name_client_replaced)
from engine import Engine, XML_TAG, NSM_NAME

_logger = logging.getLogger(__name__)
_log_handler = logging.StreamHandler()
_log_handler.setFormatter(logging.Formatter(
    f"%(name)s - %(levelname)s - %(message)s"))
_logger.setLevel(logging.WARNING)
_logger.addHandler(_log_handler)

engine = Engine()
brothers_dict = dict[str, str]()
connection_list = list[tuple[str, str]]()
saved_connections = list[tuple[str, str]]()
to_disc_connections = list[tuple[str, str]]()
jack_ports = dict[PortMode, list[JackPort]]()
for port_mode in (PortMode.NULL, PortMode.INPUT, PortMode.OUTPUT):
    jack_ports[port_mode] = list[JackPort]()

timer_dirty_check = Timer(0.300)
timer_connect_check = Timer(0.200)
nsm_server: NsmServer

def signal_handler(sig, frame):
    if sig in (signal.SIGINT, signal.SIGTERM):
        Glob.terminate = True

def set_dirty_clean():
    Glob.is_dirty = False
    nsm_server.send_dirty_state(False)

def timer_dirty_finished():
    if Glob.is_dirty:
        return

    if Glob.pending_connection:
        timer_dirty_check.start()
        return

    if is_dirty_now():
        Glob.is_dirty = True
        nsm_server.send_dirty_state(True)
    elif not Glob.dirty_state_sent:
        nsm_server.send_dirty_state(False)
        Glob.dirty_state_sent = True

def is_dirty_now() -> bool:
    for conn in connection_list:
        if not conn in saved_connections:
            # There is at least a present connection unsaved
            return True

    for sv_con in saved_connections:
        if sv_con in connection_list:
            continue

        if (sv_con[0] in [p.name for p in jack_ports[PortMode.OUTPUT]]
                and sv_con[1] in [p.name for p in jack_ports[PortMode.INPUT]]):
            # There is at least one saved connection not present
            # despite the fact its two ports are present.
            return True

    return False

def port_added(port_name: str, port_mode: int, port_type: int):
    port = JackPort()
    port.name = port_name
    port.mode = PortMode(port_mode)
    port.type = PortType(port_type)
    port.is_new = True

    jack_ports[port_mode].append(port)
    timer_connect_check.start()

def port_removed(port_name: str, port_mode: PortMode, port_type: PortType):
    for port in jack_ports[port_mode]:
        if port.name == port_name and port.type == port_type:
            jack_ports[port_mode].remove(port)
            break
    
    else:
        # strange, but in some cases,
        # JACK does not sends the good port mode at remove time.
        for pmode in (PortMode.INPUT, PortMode.OUTPUT):
            if pmode is port_mode:
                continue
            
            for port in jack_ports[pmode]:
                if port.name == port_name and port.type == port_type:
                    jack_ports[pmode].remove(port)
                    break
            break

def port_renamed(old_name: str, new_name: str,
                 port_mode: PortMode, port_type: PortType):
    for port in jack_ports[port_mode]:
        if port.name == old_name and port.type == port_type:
            port.name = new_name
            port.is_new = True
            timer_connect_check.start()
            break
    
def connection_added(port_str_a: str, port_str_b: str):
    connection_list.append((port_str_a, port_str_b))

    if Glob.pending_connection:
        may_make_one_connection()

    if (port_str_a, port_str_b) not in saved_connections:
        timer_dirty_check.start()
        
def connection_removed(port_str_a: str, port_str_b: str):
    if (port_str_a, port_str_b) in connection_list:
        connection_list.remove((port_str_a, port_str_b))

    if to_disc_connections:
        may_make_one_connection()

    timer_dirty_check.start()

def may_make_one_connection():
    if Glob.allow_disconnections:
        if to_disc_connections:
            for to_disc_con in to_disc_connections:
                if to_disc_con in connection_list:
                    _logger.info(f'disconnect ports: {to_disc_con}')
                    engine.disconnect_ports(*to_disc_con)
                    return
            else:
                to_disc_connections.clear()

    output_ports = [p.name for p in jack_ports[PortMode.OUTPUT]]
    input_ports = [p.name for p in jack_ports[PortMode.INPUT]]
    new_output_ports = [p.name for p in jack_ports[PortMode.OUTPUT] if p.is_new]
    new_input_ports = [p.name for p in jack_ports[PortMode.INPUT] if p.is_new]

    one_connected = False

    for sv_con in saved_connections:
        if (not sv_con in connection_list
                and sv_con[0] in output_ports
                and sv_con[1] in input_ports
                and (sv_con[0] in new_output_ports
                     or sv_con[1] in new_input_ports)):
            if one_connected:
                Glob.pending_connection = True
                break

            _logger.info(f'connect ports: {sv_con}')
            engine.connect_ports(*sv_con)
            one_connected = True
    else:
        Glob.pending_connection = False

        for port_mode in (PortMode.INPUT, PortMode.OUTPUT):
            for port in jack_ports[port_mode]:
                port.is_new = False

# ---- NSM callbacks ----

def open_file(project_path: str, session_name: str,
              full_client_id: str) -> tuple[Err, str]:
    saved_connections.clear()

    file_path = project_path + '.xml'
    Glob.file_path = file_path
    _logger.info(f'open file: {file_path}')

    if os.path.isfile(file_path):
        try:
            tree = ET.parse(file_path)
        except:
            _logger.error(f'unable to read file {file_path}')
            return (Err.BAD_PROJECT, f'{file_path} is not a correct .xml file')
        
        # read the DOM
        root = tree.getroot()
        if root.tag != XML_TAG:
            _logger.error(f'{file_path} is not a {XML_TAG} .xml file')
            return (Err.BAD_PROJECT, f'{file_path} is not a {XML_TAG} .xml file')
        
        graph_ports = dict[PortMode, list[str]]()
        for port_mode in (PortMode.INPUT, PortMode.OUTPUT):
            graph_ports[port_mode] = list[str]()
        
        for child in root:
            if child.tag == 'connection':
                port_from: str = child.attrib.get('from')
                port_to: str = child.attrib.get('to')
                nsm_client_from: str = child.attrib.get('nsm_client_from')
                nsm_client_to: str = child.attrib.get('nsm_client_to')

                if not port_from and port_to:
                    _logger.warning(
                        f"{debug_conn_str((port_from, port_to))} is incomplete.")
                    continue

                # ignore connection if NSM client has been definitely removed
                if Glob.monitor_states_done is MonitorStates.DONE:
                    if nsm_client_from and not nsm_client_from in brothers_dict.keys():
                        _logger.info(
                            f"{debug_conn_str((port_from, port_to))} is removed "
                            f"because NSM client {nsm_client_from} has been removed")
                        continue
                    if nsm_client_to and not nsm_client_to in brothers_dict.keys():
                        _logger.info(
                            f"{debug_conn_str((port_from, port_to))} is removed "
                            f"because NSM client {nsm_client_to} has been removed")
                        continue
                    
                saved_connections.append((port_from, port_to))
        
            elif child.tag == 'graph':
                for gp in child:
                    if gp.tag != 'group':
                        continue
                    gp_name = gp.attrib['name']
                    for pt in gp:
                        if pt.tag == 'out_port':
                            graph_ports[PortMode.OUTPUT].append(
                                ':'.join((gp_name, pt.attrib['name'])))
                        elif pt.tag == 'in_port':
                            graph_ports[PortMode.INPUT].append(
                                ':'.join((gp_name, pt.attrib['name'])))

        # re-declare all ports as new in case we are switching session
        for port_mode in (PortMode.INPUT, PortMode.OUTPUT):
            for port in jack_ports[port_mode]:
                port.is_new = True

        to_disc_connections.clear()
        # disconnect connections not existing at last save
        # if their both ports were present in the graph.
        for conn in connection_list:
            if (conn not in saved_connections
                    and conn[0] in graph_ports[PortMode.OUTPUT]
                    and conn[1] in graph_ports[PortMode.INPUT]):
                to_disc_connections.append(conn)

        if Glob.open_done_once:
            Glob.allow_disconnections = True

        may_make_one_connection()

    Glob.is_dirty = False
    Glob.open_done_once = True
    timer_dirty_check.start()
    return (Err.OK, '')

def save_file():
    if not Glob.file_path:
        return

    for connection in connection_list:
        if not connection in saved_connections:
            saved_connections.append(connection)

    # delete from saved connected all connections when there ports are present
    # and not currently connected    
    del_list = list[tuple[str, str]]()

    for sv_con in saved_connections:
        if (not sv_con in connection_list
                and sv_con[0] in [p.name for p in jack_ports[PortMode.OUTPUT]]
                and sv_con[1] in [p.name for p in jack_ports[PortMode.INPUT]]):
            del_list.append(sv_con)
            
    for del_con in del_list:
        saved_connections.remove(del_con)

    # write the XML file
    root = ET.Element(XML_TAG)
    for sv_con in saved_connections:
        conn_el = ET.SubElement(root, 'connection')
        conn_el.attrib['from'], conn_el.attrib['to'] = sv_con
        jack_con_from_name = sv_con[0].partition(':')[0].partition('/')[0]
        jack_con_to_name = sv_con[1].partition(':')[0].partition('/')[0]
        for key, value in brothers_dict.items():
            if jack_con_from_name == value:
                conn_el.attrib['nsm_client_from'] = key
            if jack_con_to_name == value:
                conn_el.attrib['nsm_client_to'] = key

    graph = ET.SubElement(root, 'graph')
    group_names = dict[str, ET.Element]()

    for port_mode in (PortMode.INPUT, PortMode.OUTPUT):
        el_name = 'in_port' if port_mode is PortMode.INPUT else 'out_port'

        for jack_port in jack_ports[port_mode]:
            gp_name, colon, port_name = jack_port.name.partition(':')
            if group_names.get(gp_name) is None:
                group_names[gp_name] = ET.SubElement(graph, 'group')
                group_names[gp_name].attrib['name'] = gp_name

                gp_base_name = gp_name.partition('/')[0]
                for key, value in brothers_dict.items():
                    if value == gp_base_name:
                        group_names[gp_name].attrib['nsm_client'] = key
                        break

            out_port_el = ET.SubElement(group_names[gp_name], el_name)
            out_port_el.attrib['name'] = port_name
    
    if sys.version_info >= (3, 9):
        # we can indent the xml tree since python3.9
        ET.indent(root, space='  ', level=0)

    _logger.info(f'save file: {Glob.file_path}')
    tree = ET.ElementTree(root)
    try:
        tree.write(Glob.file_path)
    except:
        _logger.error(f'unable to write file {Glob.file_path}')
        Glob.terminate = True
        return

    set_dirty_clean()
    return (Err.OK, 'Done')

def monitor_client_state(client_id: str, jack_name: str, is_started: int):
    if Glob.monitor_states_done is not MonitorStates.UPDATING:
        brothers_dict.clear()

    Glob.monitor_states_done = MonitorStates.UPDATING
    
    if client_id:
        brothers_dict[client_id] = jack_name
        
        if (Glob.client_changing_id is not None
                and Glob.client_changing_id[1] == client_id):
            # we are here only in the case a client id was just changed
            # we modify the saved connections in consequence.
            # Note that this client can't be started.
            ex_jack_name = Glob.client_changing_id[0]
            rm_conns = list[tuple[str, str]]()
            new_conns = list[tuple[str, str]]()

            for conn in saved_connections:
                out_port, in_port = conn
                if (port_belongs_to_client(out_port, ex_jack_name)
                        or port_belongs_to_client(in_port, ex_jack_name)):
                    rm_conns.append(conn)
                    new_conns.append(
                        (port_name_client_replaced(
                            out_port, ex_jack_name, jack_name),
                         port_name_client_replaced(
                             in_port, ex_jack_name, jack_name)))

            for conn in rm_conns:
                saved_connections.remove(conn)
            saved_connections.extend(new_conns)
            Glob.client_changing_id = None
        
    else:
        n_clients = is_started
        if len(brothers_dict) != n_clients:
            _logger.warning('list of monitored clients is incomplete !')
            ## following line would be the most obvious thing to do,
            ## but in case of problem, we could have an infinite messages loop 
            # nsm_server.send_monitor_reset()
            return

        Glob.monitor_states_done = MonitorStates.DONE

def monitor_client_event(client_id: str, event: str):
    if event == 'removed':
        if client_id in brothers_dict:
            jack_client_name = brothers_dict.pop(client_id)
        else:
            return
        
        # remove all saved connections from and to its client
        conns_to_unsave = list[tuple[str, str]]()
            
        for conn in saved_connections:
            out_port, in_port = conn
            if (port_belongs_to_client(out_port, jack_client_name)
                    or port_belongs_to_client(in_port, jack_client_name)):
                conns_to_unsave.append(conn)
        
        for conn in conns_to_unsave:
            saved_connections.remove(conn)

    elif event.startswith('id_changed_to:'):
        if client_id in brothers_dict.keys():
            Glob.client_changing_id = (
                brothers_dict[client_id], event.partition(':')[2])
        nsm_server.send_monitor_reset()

def monitor_client_updated(client_id: str, jack_name: str,
                           is_started: int):
    brothers_dict[client_id] = jack_name

def session_is_loaded():
    Glob.allow_disconnections = True
    may_make_one_connection()

# --- end of NSM callbacks --- 

def run():
    global nsm_server

    # set log level with exec arguments
    if len(sys.argv) > 1:
        read_log_level = False
        log_level = logging.WARNING

        for arg in sys.argv[1:]:
            if arg in ('-log', '--log'):
                read_log_level = True
                log_level = logging.DEBUG

            elif read_log_level:
                if arg.isdigit():
                    log_level = int(uarg)
                else:
                    uarg = arg.upper()
                    if (uarg in logging.__dict__.keys()
                            and isinstance(logging.__dict__[uarg], int)):
                        log_level = logging.__dict__[uarg]
        _logger.setLevel(log_level)

    nsm_url = os.getenv('NSM_URL')
    if not nsm_url:
        _logger.error('Could not register as NSM client.')
        sys.exit(1)

    try:
        daemon_address = liblo.Address(nsm_url)
    except:
        _logger.error('NSM_URL seems to be invalid.')
        sys.exit(1)

    if not engine.init():
        sys.exit(2)

    nsm_server = NsmServer(daemon_address)
    nsm_server.set_callback(NsmCallback.OPEN, open_file)
    nsm_server.set_callback(NsmCallback.SAVE, save_file)
    nsm_server.set_callback(
        NsmCallback.MONITOR_CLIENT_STATE, monitor_client_state)
    nsm_server.set_callback(
        NsmCallback.MONITOR_CLIENT_EVENT, monitor_client_event)
    nsm_server.set_callback(
        NsmCallback.MONITOR_CLIENT_UPDATED, monitor_client_updated)
    nsm_server.set_callback(
        NsmCallback.SESSION_IS_LOADED, session_is_loaded)
    nsm_server.announce(
        NSM_NAME, ':dirty:switch:monitor:', sys.argv[0].rpartition('/')[2])
    
    #connect program interruption signals
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    engine.fill_ports_and_connections(jack_ports, connection_list)
    
    jack_stopped = False

    while True:
        if Glob.terminate:
            break

        nsm_server.recv(50)
        for event, args in EventHandler.new_events():
            if event is Event.PORT_ADDED:
                port_added(*args)
            elif event is Event.PORT_REMOVED:
                port_removed(*args)
            elif event is Event.PORT_RENAMED:
                port_renamed(*args)
            elif event is Event.CONNECTION_ADDED:
                connection_added(*args)
            elif event is Event.CONNECTION_REMOVED:
                connection_removed(*args)
            elif event is Event.JACK_STOPPED:
                jack_stopped = True
                break
        
        if timer_dirty_check.elapsed():
            timer_dirty_finished()
        
        if timer_connect_check.elapsed():
            may_make_one_connection()

    if not jack_stopped:
        engine.quit()



from PyQt5.QtGui import QShowEvent, QFontMetrics, QFont
from PyQt5.QtWidgets import QApplication, QDialog

from .base_elements import Port, PortType, JackPortFlag
from .ui.canvas_port_info import Ui_CanvasPortInfo

_translate = QApplication.translate

class CanvasPortInfoDialog(QDialog):
    def __init__(self, parent):
        QDialog.__init__(self, parent)
        self.ui = Ui_CanvasPortInfo()
        self.ui.setupUi(self)

        self._port = None
        
        self._show_alsa_props(False)

        self.ui.toolButtonRefresh.clicked.connect(
            self.update_contents)

    def _show_alsa_props(self, yesno: bool):
        for widget in (self.ui.labelAlsaClientId, 
                       self.ui.labelColonAlsaClientId,
                       self.ui.labelAlsaClientIdNum,
                       self.ui.labelAlsaPortId,
                       self.ui.labelColonAlsaPortId,
                       self.ui.labelAlsaPortIdNum):
            widget.setVisible(yesno)

    def set_port(self, port: Port):
        self._port = port
        self.update_contents()

    def update_contents(self):
        if self._port is None:
            return

        if self._port.type is PortType.AUDIO_JACK:
            port_type_str = _translate('patchbay', "Audio")
        elif self._port.type is PortType.MIDI_JACK:
            port_type_str = _translate('patchbay', "MIDI")
        elif self._port.type is PortType.MIDI_ALSA:
            port_type_str = _translate('patchbay', "ALSA")
            self._show_alsa_props(True)
        else:
            port_type_str = _translate('patchbay', 'NULL')

        flags_list = list[str]()

        dict_flag_str = {
            JackPortFlag.IS_INPUT: _translate('patchbay', 'Input'),
            JackPortFlag.IS_OUTPUT: _translate('patchbay', 'Output'),
            JackPortFlag.IS_PHYSICAL: _translate('patchbay', 'Physical'),
            JackPortFlag.CAN_MONITOR: _translate('patchbay', 'Monitor'),
            JackPortFlag.IS_TERMINAL: _translate('patchbay', 'Terminal'),
            JackPortFlag.IS_CONTROL_VOLTAGE: _translate('patchbay', 'Control Voltage')}

        for key in dict_flag_str.keys():
            if self._port.flags & key:
                flags_list.append(dict_flag_str[key])

        port_flags_str = ' | '.join(flags_list)

        port_full_name = self._port.full_name
        if self._port.type is PortType.MIDI_ALSA:
            splitted_names = port_full_name.split(':')
            
            port_full_name = ':'.join(splitted_names[4:])
            self.ui.labelAlsaClientIdNum.setText(splitted_names[2])
            self.ui.labelAlsaPortIdNum.setText(splitted_names[3])
            self.ui.labelJackUuid.setVisible(False)
            self.ui.labelColonJackUuid.setVisible(False)
            self.ui.lineEditUuid.setVisible(False)
            
        self.ui.lineEditFullPortName.setText(port_full_name)
        self.ui.lineEditUuid.setText(str(self._port.uuid))
        self.ui.labelPortType.setText(port_type_str)
        self.ui.labelPortFlags.setText(port_flags_str)
        self.ui.labelPrettyName.setText(self._port.pretty_name)
        self.ui.labelPortOrder.setText(str(self._port.order))
        self.ui.labelPortGroup.setText(self._port.mdata_portgroup)

        self.ui.groupBoxMetadatas.setVisible(bool(
            self._port.pretty_name
            or self._port.order is not None
            or self._port.mdata_portgroup))
    
    def showEvent(self, event: QShowEvent) -> None:
        self.resize(0, 0)
        self.ui.lineEditFullPortName.setMinimumWidth(
            QFontMetrics(QFont()).width(
            self.ui.lineEditFullPortName.text()) + 20
        )
        super().showEvent(event)


from typing import Callable
from ctypes import (
    CFUNCTYPE,
    POINTER,
    c_void_p,
    c_int,
    c_char_p,
    CDLL
)

from .types import (
    jack_client_t,
    jack_status_t,
    jack_nframes_t,
    jack_port_id_t,
    jack_latency_callback_mode_t,
    jack_transport_state_t,
    jack_position_t,
    jack_session_event_t,
    jack_uuid_t,
    jack_property_change_t
)


class _Cb:
    def __init__(self, ref: str, callback: CFUNCTYPE, restype, suffix='_callback'):
        self.ref = ref
        self.callback = callback
        self.restype = restype
        self.setter_name = ref + suffix
        try:
            func = getattr(globals()['_jlib'], 'jack_' + self.setter_name)
            func.argstypes = [POINTER(jack_client_t), self.callback, c_void_p]
            func.restype = self.restype
            self.jlib_func = func
        except AttributeError:
            self.jlib_func = None


_used_callbacks = list[CFUNCTYPE]()
_jlib = None
_cbs = tuple[_Cb]()

def init_callback_setter(jlib: CDLL):
    global _jlib, _cbs
    _jlib = jlib
    _cbs = (
        _Cb('set_thread_init',
            CFUNCTYPE(None, c_void_p),
            c_int),
        _Cb('on_shutdown',
            CFUNCTYPE(None, c_void_p),
            None,
            suffix=''),
        _Cb('on_info_shutdown',
            CFUNCTYPE(None, jack_status_t, c_char_p, c_void_p),
            None,
            suffix=''),
        _Cb('set_process',
            CFUNCTYPE(c_int, jack_nframes_t, c_void_p),
            c_int),
        _Cb('set_freewheel',
            CFUNCTYPE(None, c_int, c_void_p),
            c_int),
        _Cb('set_buffer_size',
            CFUNCTYPE(c_int, jack_nframes_t, c_void_p),
            c_int),
        _Cb('set_sample_rate',
            CFUNCTYPE(c_int, jack_nframes_t, c_void_p),
            c_int),
        _Cb('set_client_registration',
            CFUNCTYPE(None, c_char_p, c_int, c_void_p),
            c_int),
        _Cb('set_client_rename',
            CFUNCTYPE(c_int, c_char_p, c_char_p, c_void_p),
            c_int),
        _Cb('set_port_registration',
            CFUNCTYPE(None, jack_port_id_t, c_int, c_void_p),
            c_int),
        _Cb('set_port_connect',
            CFUNCTYPE(None, jack_port_id_t, jack_port_id_t, c_int, c_void_p),
            c_int),
        _Cb('set_port_rename',
            CFUNCTYPE(None, jack_port_id_t, c_char_p, c_char_p, c_void_p),
            c_int),
        _Cb('set_graph_order',
            CFUNCTYPE(c_int, c_void_p),
            c_int),
        _Cb('set_xrun',
            CFUNCTYPE(c_int, c_void_p),
            c_int),
        _Cb('set_latency',
            CFUNCTYPE(None, jack_latency_callback_mode_t, c_void_p),
            c_int),
        # other callbacks
        _Cb('set_sync',
            CFUNCTYPE(c_int, jack_transport_state_t, POINTER(jack_position_t), c_void_p),
            c_int),
        _Cb('set_session',
            CFUNCTYPE(None, POINTER(jack_session_event_t), c_void_p),
            c_int),
        _Cb('set_property_change',
            CFUNCTYPE(None, jack_uuid_t, c_char_p, jack_property_change_t, c_void_p),
            c_int),
        ## following is not really a callback setter, but it uses the same scheme
        # _Cb('set_process_thread',
        #     CFUNCTYPE(c_void_p, c_void_p),
        #     c_int,
        #     suffix='')
        )


def callback_setter(func: Callable):
    ''' decorator for callback setter.
        note that the decorated function is never executed
        everything depends on its name. '''
    for _cb in _cbs:
        if func.__name__ == _cb.setter_name:
            break
    else:
        raise BaseException(
            f'callback_setter decorator: no _Cb found with setter_name "{func.__name__}"')
    
    def wrapper(client, callback, arg):
        if _cb.jlib_func is None:
            if _cb.restype is c_int:
                return -1
            return None
        
        _callback = _cb.callback(callback)
        _used_callbacks.append(_callback)
        return _cb.jlib_func(client, _callback, arg)
    return wrapper


from enum import Enum
import logging
import os
import shutil
import subprocess
from typing import TYPE_CHECKING, Callable, Union

from PyQt5.QtCore import QProcess, QTimer
from server_sender import ServerSender
import ray

if TYPE_CHECKING:
    from session import Session

_logger = logging.getLogger(__name__)


class CopyState(Enum):
    OFF = 0
    COPYING = 1
    DONE = 2


class CopyFile:
    orig_path = ""
    dest_path = ""
    state = CopyState.OFF
    size = 0


class FileCopier(ServerSender):
    def __init__(self, session: 'Session'):
        ServerSender.__init__(self)
        self.session = session

        self._client_id = ''
        self._src_is_factory = False
        self._next_function = None
        self._abort_function = None
        self._next_args = []
        self._copy_files = list[CopyFile]()
        self._copy_size = 0
        self._aborted = False
        self._is_active = False

        self._process = QProcess()
        self._process.finished.connect(self._process_finished)
        if ray.QT_VERSION >= (5, 6):
            self._process.errorOccurred.connect(self._error_occurred)

        self._timer = QTimer()
        self._timer.setInterval(250)
        self._timer.setSingleShot(True)
        self._timer.timeout.connect(self._check_progress_size)

        self._abort_src_addr = None
        self._abort_src_path = ''

    def _get_file_size(self, filepath) -> int:
        if not os.path.exists(filepath):
            return 0

        try:
            du_full = subprocess.check_output(
                ['nice', '-n', '15', 'du', '-sb', filepath]).decode()
        except BaseException as e:
            _logger.error(str(e))
            _logger.error(f'unable to decode size of file {filepath}')
            du_full = ""

        if not du_full:
            return 0

        du_str = du_full.split('\t')[0]
        if not du_str.isdigit():
            return 0

        return int(du_str)

    def _check_progress_size(self):
        current_size = 0
        self._timer.stop()

        for copy_file in self._copy_files:
            if copy_file.state is CopyState.DONE:
                current_size += copy_file.size
            elif copy_file.state is CopyState.COPYING:
                current_size += self._get_file_size(copy_file.dest_path)
                break

        if current_size and self._copy_size:
            progress = float(current_size/self._copy_size)

            if self._client_id:
                self.send_gui('/ray/gui/client/progress',
                              self._client_id, progress)
            elif self.session.session_id:
                self.send_gui('/ray/gui/server/parrallel_copy_progress',
                              self.session.session_id, progress)
            else:
                self.send_gui('/ray/gui/server/progress', progress)

            self.session.osc_reply('/ray/net_daemon/duplicate_state', progress)

        self._timer.start()

    def _process_finished(self, exit_code: int, exit_status: int):
        self._timer.stop()

        for copy_file in self._copy_files:
            if copy_file.state is CopyState.COPYING:
                copy_file.state = CopyState.DONE
                if self._src_is_factory:
                    try:
                        subprocess.run(['chmod', '-R', '+w',
                                        copy_file.dest_path])
                    except BaseException as e:
                        _logger.error(
                            f'Failed to set path writable {copy_file.dest_path}\n'
                            f'{str(e)}')
                break

        if self._aborted:
            ##remove all created files
            for copy_file in self._copy_files:
                if copy_file.state is not CopyState.OFF:
                    file_to_remove = copy_file.dest_path

                    if os.path.exists(file_to_remove):
                        try:
                            if os.path.isfile(file_to_remove):
                                os.remove(file_to_remove)
                            elif os.path.isdir(file_to_remove):
                                shutil.rmtree(file_to_remove)
                        except:
                            if self._abort_src_addr and self._abort_src_path:
                                self.send(self._abort_src_addr,
                                          '/error_minor',
                                          self._abort_src_path,
                                          ray.Err.SUBPROCESS_CRASH,
                                          "%s hasn't been removed !")

            self._is_active = False
            self._send_copy_state_to_gui(0)
            self._abort_function(*self._next_args)
            return

        # run next_function if copy is terminated
        for copy_file in self._copy_files:
            if copy_file.state is not CopyState.DONE:
                break
        else:
            self._is_active = False
            self._send_copy_state_to_gui(0)

            if self._next_function:
                self._next_function(*self._next_args)

            return

        self._next_process()

    def _error_occurred(self):
        #todo make something else
        self._process_finished(0, 0)

    def _next_process(self):
        self._is_active = True

        for copy_file in self._copy_files:
            if copy_file.state is CopyState.OFF:
                copy_file.state = CopyState.COPYING
                self._process.start(
                    'nice',
                    ['-n', '+15', 'cp', '-R',
                     copy_file.orig_path, copy_file.dest_path])
                break

        self._timer.start()

    def _start(self, src_list: Union[str, list[str]], dest_dir: str,
               next_function: Callable, abort_function: Callable,
               next_args=[]):
        self._abort_function = abort_function
        self._next_function = next_function
        self._next_args = next_args

        self._aborted = False
        self._copy_size = 0
        self._copy_files.clear()

        dest_path_exists = bool(os.path.exists(dest_dir))
        if dest_path_exists:
            if not os.path.isdir(dest_dir):
                #TODO send error, but it should not append
                self._abort_function(*self._next_args)
                return

        if isinstance(src_list, str):
            src_dir = src_list
            src_list = list[str]()

            if not os.path.isdir(src_dir):
                self._abort_function(*self._next_args)
                return

            try:
                tmp_list = os.listdir(src_dir)
            except:
                self._abort_function(*self._next_args)
                return

            for path in tmp_list:
                if path == '.ray-snapshots':
                    continue

                full_path = "%s/%s" % (src_dir, path)
                src_list.append(full_path)

            if not dest_path_exists:
                try:
                    os.makedirs(dest_dir)
                except:
                    self._abort_function(*self._next_args)
                    return

        for orig_path in src_list:
            copy_file = CopyFile()
            copy_file.state = CopyState.OFF
            copy_file.orig_path = orig_path
            copy_file.size = self._get_file_size(orig_path)

            self._copy_size += copy_file.size

            if dest_path_exists:
                copy_file.dest_path = "%s/%s" % (dest_dir,
                                                 os.path.basename(orig_path))
            else:
                #WARNING works only with one file !!!
                copy_file.dest_path = dest_dir

            self._copy_files.append(copy_file)

        if self._copy_files:
            self._send_copy_state_to_gui(1)
            self._next_process()
        else:
            self._next_function(*self._next_args)

    def _send_copy_state_to_gui(self, state:int):
        if self.session.session_id:
            self.send_gui('/ray/gui/server/parrallel_copy_state',
                          self.session.session_id, state)
        else:
            self.send_gui('/ray/gui/server/copying', state)

    def start_client_copy(
            self, client_id: str, src_list: list[str], dest_dir: str,
            next_function: Callable, abort_function: Callable,
            next_args=[], src_is_factory=False):
        self._client_id = client_id
        self._src_is_factory = src_is_factory
        self._start(src_list, dest_dir, next_function,
                    abort_function, next_args)

    def start_session_copy(
            self, src_dir: str, dest_dir: str,
            next_function: Callable, abort_function: Callable, next_args=[],
            src_is_factory=False):
        self._client_id = ''
        self._src_is_factory = src_is_factory
        self._start(src_dir, dest_dir, next_function,
                     abort_function, next_args)

    def abort(self, abort_function=None, next_args=[]):
        if abort_function:
            self._abort_function = abort_function
            self._next_args = next_args

        self._timer.stop()

        if self._process.state() == QProcess.Running:
            self._aborted = True
            self._process.terminate()

    def is_active(self, client_id=''):
        if client_id and client_id != self._client_id:
            return False

        return self._is_active



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys

import jacklib
from jacklib.helpers import c_char_p_p_to_list, get_jack_status_error_string


status = jacklib.jack_status_t()
client = jacklib.client_open("list-port-info", jacklib.JackOptions.NO_START_SERVER, status)
err = get_jack_status_error_string(status)

if status.value:
    if status.value & jacklib.JackStatus.NAME_NOT_UNIQUE:
        print("Non-fatal JACK status: %s" % err, file=sys.stderr)
    elif status.value & jacklib.JackStatus.SERVER_STARTED:
        # Should not happen, since we use the JackOptions.NO_START_SERVER option
        print("Unexpected JACK status: %s" % err, file=sys.stderr)
    else:
        sys.exit("Error connecting to JACK server: %s" % err)

for portname in c_char_p_p_to_list(jacklib.get_ports(client)):
    port = jacklib.port_by_name(client, portname)
    uuid = jacklib.port_uuid(port)

    print("Port name: %s\nUUID: %s" % (portname, uuid))
    num_aliases, *aliases = jacklib.port_get_aliases(port)
    if num_aliases:
        print("Aliases: %s" % ", ".join(aliases[:num_aliases]))

    pretty_name = jacklib.get_port_pretty_name(client, portname)
    if pretty_name:
        print("Pretty-name: {}".format(pretty_name))

    props = jacklib.get_port_properties(client, portname)
    if props:
        print("Properties:")
        for prop in props:
            print(" * {}: {} (type: {})".format(prop.key, prop.value, prop.type))

    print('')

jacklib.client_close(client)



"""Python bindings for libjack using ctypes."""

# Copyright (C) 2010-2020 Filipe Coelho <falktx@falktx.com>
#               2016-2022 Christopher Arndt <info@chrisarndt.de>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# For a full copy of the GNU General Public License see the file COPYING.md.

from .types import *
from .enums import *
from .api import *  # noqa
from .version import __version__  # noqa



import argparse
import os
import sys
from dataclasses import dataclass
from typing import TYPE_CHECKING
from pathlib import Path

from PyQt5.QtCore import (QCoreApplication, QStandardPaths, QSettings,
                          QDateTime, QLocale)
try:
    import liblo
except ImportError:
    import pyliblo3 as liblo


import ray

if TYPE_CHECKING:
    from client import Client


settings = QSettings()

# remember here associated desktop files for executables
# key: executable, value: desktop_file name
exec_and_desktops = dict[str, str]()


def dirname(*args) -> str:
    return os.path.dirname(*args)

def basename(*args) -> str:
    return os.path.basename(*args)

def get_app_config_path() -> str:
    return "%s/%s" % (
            QStandardPaths.writableLocation(QStandardPaths.ConfigLocation),
            QCoreApplication.organizationName())

def get_code_root() -> Path:
    return Path(__file__).parent.parent.parent

def is_pid_child_of(child_pid, parent_pid) -> bool:
    if child_pid < parent_pid:
        return False

    ppid = child_pid

    while ppid > parent_pid:
        try:
            proc_file = open('/proc/%i/status' % ppid, 'r')
            proc_contents = proc_file.read()
        except BaseException:
            return False

        for line in proc_contents.split('\n'):
            if line.startswith('PPid:'):
                ppid_str = line.rpartition('\t')[2]
                if ppid_str.isdigit():
                    ppid = int(ppid_str)
                    break
        else:
            return False

    #while ppid != parent_pid and ppid > 1 and ppid != this_pid:
        #try:
            #ppid = int(subprocess.check_output(
                #['ps', '-o', 'ppid=', '-p', str(ppid)]))
        #except BaseException:
            #return False

    if ppid == parent_pid:
        return True

    return False

def highlight_text(string) -> str:
    if "'" in string:
        return '"%s"' % string
    return "'%s'" % string

def init_daemon_tools():
    if CommandLineArgs.config_dir:
        l_settings = QSettings(CommandLineArgs.config_dir)
    else:
        l_settings = QSettings()

    RS.set_settings(l_settings)

    RS.set_non_active_clients(
        ray.get_list_in_settings(l_settings, 'daemon/non_active_list'))
    RS.set_favorites(ray.get_list_in_settings(l_settings, 'daemon/favorites'))
    TemplateRoots.init_config()

def get_git_default_un_and_ignored(executable:str) -> tuple[list[str], list[str]]:
    ignored = list[str]()
    unignored = list[str]()

    if executable in ('luppp', 'sooperlooper', 'sooperlooper_nsm'):
        unignored.append('.wav')

    elif executable == 'samplv1_jack':
        unignored = ['.wav', '.flac', '.ogg', '.mp3']

    return (ignored, unignored)


@dataclass
class AppTemplate:
    template_name: str
    template_client: 'Client'
    display_name: str
    templates_root: str


class RS:
    settings = QSettings()
    non_active_clients = []
    favorites = list[ray.Favorite]()

    @classmethod
    def set_settings(cls, settings):
        del cls.settings
        cls.settings = settings

    @classmethod
    def set_non_active_clients(cls, nalist):
        del cls.non_active_clients
        cls.non_active_clients = nalist

    @classmethod
    def set_favorites(cls, favorites: list[ray.Favorite]):
        cls.favorites.clear()
        
        for fav in favorites:
            fav_dict = fav.__dict__
            if (isinstance(fav_dict.get('name'), str)
                    and isinstance(fav_dict.get('factory'), bool)
                    and isinstance(fav_dict.get('icon'), str)):
                display_name = fav_dict.get('display_name')
                if not isinstance(display_name, str):
                    display_name = ''
                
                if not display_name:
                    display_name = fav_dict.get('name')
                cls.favorites.append(ray.Favorite(
                    fav_dict.get('name'),
                    fav_dict.get('icon'),
                    fav_dict.get('factory'),
                    display_name))

class TemplateRoots:
    net_session_name = ".ray-net-session-templates"
    factory_sessions = get_code_root() / 'session_templates'
    factory_clients = get_code_root() / 'client_templates'    
    factory_clients_xdg = Path('/etc/xdg/raysession/client_templates')

    @classmethod
    def init_config(cls):
        if CommandLineArgs.config_dir:
            app_config_path = CommandLineArgs.config_dir
        else:
            app_config_path = get_app_config_path()

        cls.user_sessions = Path(app_config_path) / 'session_templates'
        cls.user_clients = Path(app_config_path) / 'client_templates'


class Terminal:
    _last_client_name = ''

    @classmethod
    def message(cls, string, server_port=0):
        if cls._last_client_name and cls._last_client_name != 'daemon':
            sys.stderr.write('\n')

        sys.stderr.write('[\033[90mray-daemon\033[0m]\033[92m%s\033[0m\n'
                            % string)

        log_dir = "%s/logs" % get_app_config_path()
        if server_port:
            log_file_path = "%s/%i" % (log_dir, server_port)
        else:
            log_file_path = "%s/dummy" % log_dir

        if not os.path.exists(log_dir):
            os.makedirs(log_dir)

        log_file = open(log_file_path, 'a')

        date_time = QDateTime.currentDateTime()
        locale = QLocale(QLocale.English)
        date_format = locale.toString(date_time, "ddd MMM d hh:mm:ss yyyy")

        log_file.write("%s: %s\n" % (date_format, string))

        cls._last_client_name = 'daemon'

    @classmethod
    def snapshoter_message(cls, byte_string, command=''):
        snapshoter_str = "snapshoter:.%s" % command

        if cls._last_client_name != snapshoter_str:
            sys.stderr.write('\n[\033[90mray-daemon-git%s\033[0m]\n'
                             % command)
        sys.stderr.buffer.write(byte_string)

        cls._last_client_name = snapshoter_str

    @classmethod
    def scripter_message(cls, byte_string, command=''):
        scripter_str = "scripter:.%s" % command

        if cls._last_client_name != scripter_str:
            sys.stderr.write('\n[\033[90mray-daemon %s script\033[0m]\n'
                             % command)
        sys.stderr.buffer.write(byte_string)

        cls._last_client_name = scripter_str

    @classmethod
    def client_message(cls, byte_string, client_name, client_id):
        client_str = "%s.%s" % (client_name, client_id)

        if (not CommandLineArgs.debug_only
                and not CommandLineArgs.no_client_messages):
            if cls._last_client_name != client_str:
                sys.stderr.write('\n[\033[90m%s-%s\033[0m]\n'
                                    % (client_name, client_id))
            sys.stderr.buffer.write(byte_string)

        cls._last_client_name = client_str

    @classmethod
    def warning(cls, string):
        sys.stderr.write('[\033[90mray-daemon\033[0m]%s\033[0m\n' % string)
        cls._last_client_name = 'daemon'


class CommandLineArgs(argparse.Namespace):
    session_root = ''
    hidden = False
    osc_port = 0
    findfreeport = True
    control_url: liblo.Address = None
    gui_url: liblo.Address = None
    gui_port: liblo.Address = 0
    gui_pid = 0
    config_dir = ''
    debug = False
    debug_only = False
    no_client_messages = False
    session = ''
    no_options = False

    @classmethod
    def eat_attributes(cls, parsed_args):
        for attr_name in dir(parsed_args):
            if not attr_name.startswith('_'):
                setattr(cls, attr_name, getattr(parsed_args, attr_name))

        if cls.debug_only:
            cls.debug = True

        if cls.osc_port == 0:
            cls.osc_port = 16187
            cls.findfreeport = True

        if cls.config_dir and not os.access(cls.config_dir, os.W_OK):
            sys.stderr.write(
                '%s is not a writable config dir, try another one\n'
                    % cls.config_dir)
            sys.exit(1)

class ArgParser(argparse.ArgumentParser):
    def __init__(self):
        argparse.ArgumentParser.__init__(self)
        _translate = QCoreApplication.translate

        default_root = "%s/%s" % (
            os.getenv('HOME'),
            _translate('daemon', 'Ray Network Sessions'))

        self.add_argument('--session-root', '-r', type=str,
                          default=default_root,
                          help='set root folder for sessions')
        self.add_argument('--session', '-s', type=str, default='',
                          help='session to load at startup')
        self.add_argument('--osc-port', '-p',
                          type=int, default=0,
                          help='select OSC port for the daemon')
        self.add_argument('--findfreeport', action='store_true',
                          help='find another port if port is not free')
        self.add_argument('--gui-url', type=ray.get_liblo_address,
                          help=argparse.SUPPRESS)
        self.add_argument('--gui-port', type=ray.get_liblo_address_from_port,
                          help=argparse.SUPPRESS)
        self.add_argument('--gui-pid', type=int,
                          help=argparse.SUPPRESS)
        self.add_argument('--control-url', type=ray.get_liblo_address,
                          help=argparse.SUPPRESS)
        self.add_argument('--no-options', action='store_true',
            help='start without any option and do not save options at quit')
        self.add_argument('--hidden', action='store_true',
            help='hide for ray_control')
        self.add_argument('--config-dir', '-c', type=str, default='',
                          help='use a custom config dir')
        self.add_argument('--debug', '-d', action='store_true',
                          help='see all OSC messages')
        self.add_argument('--debug-only', '-do', action='store_true',
                          help='debug without client messages')
        self.add_argument('--no-client-messages', '-ncm', action='store_true',
                          help='do not print client messages')

        self.add_argument('-v', '--version', action='version',
                          version=ray.VERSION)

        parsed_args = argparse.ArgumentParser.parse_args(self)
        CommandLineArgs.eat_attributes(parsed_args)




import logging

import jacklib
from jacklib.helpers import c_char_p_p_to_list
from jacklib.api import JackPortFlags, JackOptions, pointer, jack_client_t

import jack_callbacks

from bases import JackPort, PortMode, ProtoEngine, PortType

_logger = logging.getLogger(__name__)


class JackEngine(ProtoEngine):
    def __init__(self):
        super().__init__()
        self._jack_client: 'pointer[jack_client_t]' = None

    def init(self) -> bool:
        self._jack_client = jacklib.client_open(
            'ray-jackpatch',
            JackOptions.NO_START_SERVER,
            None)

        if not self._jack_client:
            _logger.error('Unable to make a jack client !')
            return False
        
        jack_callbacks.set_callbacks(self._jack_client)
        jacklib.activate(self._jack_client)
        return True

    def fill_ports_and_connections(
            self, all_ports: dict[PortMode, list[JackPort]],
            connection_list: list[tuple[str, str]]):
        '''get all current JACK ports and connections at startup'''
        port_name_list = c_char_p_p_to_list(
            jacklib.get_ports(self._jack_client, "", "", 0))

        for port_name in port_name_list:
            jack_port = JackPort()
            jack_port.name = port_name

            port_ptr = jacklib.port_by_name(self._jack_client, port_name)
            port_flags = jacklib.port_flags(port_ptr)

            if port_flags & JackPortFlags.IS_INPUT:
                jack_port.mode = PortMode.INPUT
            elif port_flags & JackPortFlags.IS_OUTPUT:
                jack_port.mode = PortMode.OUTPUT
            else:
                jack_port.mode = PortMode.NULL

            port_type_str = jacklib.port_type(port_ptr)
            if port_type_str == jacklib.JACK_DEFAULT_AUDIO_TYPE:
                jack_port.type = PortType.AUDIO
            elif port_type_str == jacklib.JACK_DEFAULT_MIDI_TYPE:
                jack_port.type = PortType.MIDI
            else:
                jack_port.type = PortType.NULL

            jack_port.is_new = True

            all_ports[jack_port.mode].append(jack_port)

            if jack_port.mode is PortMode.OUTPUT:
                for port_con_name in jacklib.port_get_all_connections(
                        self._jack_client, port_ptr):
                    connection_list.append((port_name, port_con_name))

    def connect_ports(self, port_out: str, port_in: str):
        jacklib.connect(self._jack_client, port_out, port_in)

    def disconnect_ports(self, port_out: str, port_in: str):
        jacklib.disconnect(self._jack_client, port_out, port_in)

    def quit(self):
        jacklib.deactivate(self._jack_client)
        jacklib.client_close(self._jack_client)
    



import json
import os
from pathlib import Path
import subprocess
import time
from typing import TYPE_CHECKING
try:
    from liblo import Address
except ImportError:
    from pyliblo3 import Address
from PyQt5.QtCore import QCoreApplication, QProcess
from PyQt5.QtXml  import QDomDocument

import ray

from client import Client
from multi_daemon_file import MultiDaemonFile
from signaler import Signaler
from daemon_tools import (Terminal, RS, dirname,
                          is_pid_child_of, highlight_text)
from session import OperatingSession
import xdg
from patch_rewriter import rewrite_jack_patch_files

_translate = QCoreApplication.translate
signaler = Signaler.instance()


def session_operation(func):
    def wrapper(*args, **kwargs):
        if len(args) < 4:
            return

        sess, path, osc_args, src_addr, *rest = args
        
        if TYPE_CHECKING:
            assert isinstance(sess, SignaledSession)
            assert isinstance(path, str)

        if sess.steps_order:
            sess.send(src_addr, "/error", path, ray.Err.OPERATION_PENDING,
                      "An operation pending.")
            return

        if sess.file_copier.is_active():
            if path.startswith('/nsm/server/'):
                sess.send(src_addr, "/error", path, ray.Err.OPERATION_PENDING,
                      "An operation pending.")
            else:
                sess.send(src_addr, "/error", path, ray.Err.COPY_RUNNING,
                        "ray-daemon is copying files.\n"
                            + "Wait copy finish or abort copy,\n"
                            + "and restart operation !\n")
            return

        sess.remember_osc_args(path, osc_args, src_addr)

        response = func(*args, **kwargs)
        sess.next_function()

        return response
    return wrapper

def client_action(func):
    def wrapper(*args, **kwargs):
        if len(args) < 4:
            return

        sess, path, osc_args, src_addr, *rest = args
        
        if TYPE_CHECKING:
            assert isinstance(sess, SignaledSession)
            assert isinstance(osc_args, list)

        client_id: str = osc_args.pop(0)

        for client in sess.clients:
            if client.client_id == client_id:
                response = func(*args, client, **kwargs)
                break
        else:
            sess.send_error_no_client(src_addr, path, client_id)
            return

        return response
    return wrapper

# There is only one possible instance of SignaledSession
# This is not the case for Session and OperatingSession.
# This session receives signals from OSC server.
class SignaledSession(OperatingSession):
    def __init__(self, root: str):
        OperatingSession.__init__(self, root)

        signaler.osc_recv.connect(self.osc_receive)
        signaler.dummy_load_and_template.connect(self.dummy_load_and_template)

        # dict where:
        # key is a sessions root folder
        # value is a list of recent sessions
        self.recent_sessions: dict[str, list[str]] = RS.settings.value(
            'daemon/recent_sessions', {}, type=dict)
        self.check_recent_sessions_existing()

        self.preview_dummy_session = None
        self.dummy_sessions = list[DummySession]()
        self._next_session_id = 1
        
        self._folder_sizes_and_dates = []
        
        self._cache_folder_sizes_path = xdg.xdg_cache_home().joinpath(
            'RaySession', 'folder_sizes.json')

        if os.path.isfile(self._cache_folder_sizes_path):
            try:
                self._folder_sizes_and_dates = json.load(
                    self._cache_folder_sizes_path)
            except:
                # cache file load failed and this is really not strong
                pass
    
    def _get_new_dummy_session_id(self)->int:
        to_return = self._next_session_id
        self._next_session_id += 1
        return to_return

    def _new_dummy_session(self, root:str):
        new_dummy = DummySession(root, self._get_new_dummy_session_id())
        self.dummy_sessions.append(new_dummy)
        return new_dummy

    def save_folder_sizes_cache_file(self):
        cache_dir = dirname(self._cache_folder_sizes_path)
        if not os.path.exists(cache_dir):
            try:
                os.makedirs(cache_dir)
            except:
                # can't save cache file, this is really not strong
                return
        
        try:
            file = open(self._cache_folder_sizes_path, 'w')
            json.dump(self._folder_sizes_and_dates, file)
            file.close()
        except:
            # cache file save failed, not strong
            pass

    def osc_receive(self, path, args, types, src_addr):
        nsm_equivs = {"/nsm/server/add" : "/ray/session/add_executable",
                      "/nsm/server/save": "/ray/session/save",
                      "/nsm/server/open": "/ray/server/open_session",
                      "/nsm/server/new" : "/ray/server/new_session",
                      "/nsm/server/duplicate": "/ray/session/duplicate",
                      "/nsm/server/close": "/ray/session/close",
                      "/nsm/server/abort": "/ray/session/abort",
                      "/nsm/server/quit" : "/ray/server/quit"}
                      # /nsm/server/list is not used here because it doesn't
                      # works as /ray/server/list_sessions
        nsm_path = nsm_equivs.get(path)
        func_path = nsm_path if nsm_path else path

        func_name = func_path.replace('/', '_')

        if func_name in self.__dir__():
            function = self.__getattribute__(func_name)
            function(path, args, src_addr)

    def send_error_no_client(self, src_addr, path, client_id):
        self.send(src_addr, "/error", path, ray.Err.CREATE_FAILED,
                  _translate('GUIMSG', "No client with this client_id:%s")
                    % client_id)

    def send_error_copy_running(self, src_addr, path):
        self.send(src_addr, "/error", path, ray.Err.COPY_RUNNING,
                  _translate('GUIMSG', "Impossible, copy running !"))

    ############## FUNCTIONS CONNECTED TO SIGNALS FROM OSC ###################

    def _nsm_server_announce(self, path: str, args, src_addr):
        client_name, capabilities, executable_path, major, minor, pid = args

        if self.wait_for == ray.WaitFor.QUIT:
            if path.startswith('/nsm/server/'):
                # Error is wrong but compatible with NSM API
                self.send(src_addr, "/error", path, ray.Err.NO_SESSION_OPEN,
                          "Sorry, but there's no session open "
                          + "for this application to join.")
            return

        # we can't be absolutely sure that the announcer is the good one
        # but if client announce a known PID,
        # we can be sure of which client is announcing

        for client in self.clients:
            if client.pid == pid and not client.nsm_active and client.is_running():
                client.server_announce(path, args, src_addr, False)
                break
        else:
            for client in self.clients:
                if (not client.nsm_active and client.is_running()
                        and is_pid_child_of(pid, client.pid)):
                    client.server_announce(path, args, src_addr, False)
                    break
            else:
                for client in self.clients:
                    if (client.launched_in_terminal
                            and client.process_drowned
                            and client.executable_path == executable_path):
                        # when launched in terminal
                        # the client process can be stopped
                        # because the terminal process is 'linked' to an existing instance
                        # then, we may can say this stopped client is the good one,
                        # and we declare it as external because we won't check its process
                        # state with QProcess.state().
                        client.server_announce(path, args, src_addr, True)
                        break
                else:
                    # Client launched externally from daemon
                    # by command : $:NSM_URL=url executable
                    client = self._new_client(executable_path)
                    self.externals_timer.start()
                    self.send_monitor_event('joined', client.client_id)
                    client.server_announce(path, args, src_addr, True)

        if self.wait_for == ray.WaitFor.ANNOUNCE:
            self.end_timer_if_last_expected(client)

    def _reply(self, path, args, src_addr):
        if self.wait_for == ray.WaitFor.QUIT:
            return

        message = args[1]
        client = self.get_client_by_address(src_addr)
        if client:
            client.set_reply(ray.Err.OK, message)

            server = self.get_server()
            if (server is not None
                    and server.server_status == ray.ServerStatus.READY
                    and server.options & ray.Option.DESKTOPS_MEMORY):
                self.desktops_memory.replace()
        else:
            self.message("Reply from unknown client")

    def _error(self, path, args, src_addr):
        path, errcode, message = args

        client = self.get_client_by_address(src_addr)
        if client:
            client.set_reply(errcode, message)

            if self.wait_for == ray.WaitFor.REPLY:
                self.end_timer_if_last_expected(client)
        else:
            self.message("error from unknown client")

    def _nsm_client_label(self, path, args, src_addr):
        client = self.get_client_by_address(src_addr)
        if client:
            client.set_label(args[0])

    def _nsm_client_network_properties(self, path, args, src_addr):
        client = self.get_client_by_address(src_addr)
        if client:
            net_daemon_url, net_session_root = args
            client.set_network_properties(net_daemon_url, net_session_root)

    def _nsm_client_no_save_level(self, path, args, src_addr):
        client = self.get_client_by_address(src_addr)
        if client and client.is_capable_of(':warning-no-save:'):
            client.no_save_level = args[0]

            self.send_gui('/ray/gui/client/no_save_level',
                           client.client_id, client.no_save_level)

    def _ray_server_ask_for_patchbay(self, path, args, src_addr):
        # if we are here, this means we need a patchbay to osc to run
        server = self.get_server()
        if server is None:
            return
        QProcess.startDetached('ray-jackpatch_to_osc',
                               [str(server.port), src_addr.url])

    def _ray_server_abort_copy(self, path, args, src_addr):
        self.file_copier.abort()

    def _ray_server_abort_parrallel_copy(self, path, args, src_addr):
        session_id = args[0]
        
        for dummy_session in self.dummy_sessions:
            if dummy_session.session_id == session_id:
                dummy_session.file_copier.abort()
                break
        self.send(src_addr, '/reply', path, 'Parrallel copy aborted')

    def _ray_server_abort_snapshot(self, path, args, src_addr):
        self.snapshoter.abort()

    def _ray_server_change_root(self, path, args, src_addr):
        session_root = args[0]
        if self.path:
            self.send(src_addr, '/error', path, ray.Err.SESSION_LOCKED,
                      "impossible to change root. session %s is loaded"
                      % self.path)
            return

        if not os.path.exists(session_root):
            try:
                os.makedirs(session_root)
            except:
                self.send(src_addr, '/error', path, ray.Err.CREATE_FAILED,
                          "invalid session root !")
                return

        if not os.access(session_root, os.W_OK):
            self.send(src_addr, '/error', path, ray.Err.CREATE_FAILED,
                      "unwriteable session root !")
            return

        self.root = session_root

        multi_daemon_file = MultiDaemonFile.get_instance()
        if multi_daemon_file:
            multi_daemon_file.update()

        self.send(src_addr, '/reply', path,
                  "root folder changed to %s" % self.root)
        self.send_gui('/ray/gui/server/root', self.root)

        if self.root not in self.recent_sessions.keys():
            self.recent_sessions[self.root] = []
        self.send_gui('/ray/gui/server/recent_sessions',
                       *self.recent_sessions[self.root])

    def _ray_server_list_client_templates(self, path, args, src_addr):
        # def send_gui_template_update(template_dict: dict):
        #     template_name = template_dict['template_name']
        #     template_client = template_dict['template_client']
        #     display_name = template_dict['display_name']
            
        #     self.send_gui(
        #         '/ray/gui/client_template_update',
        #         int(factory), template_name, display_name,
        #         *template_client.spread())
        #     if template_client.protocol == ray.Protocol.RAY_HACK:
        #         self.send_gui(
        #             '/ray/gui/client_template_ray_hack_update',
        #             int(factory), template_name,
        #             *template_client.ray_hack.spread())
        #     elif template_client.protocol == ray.Protocol.RAY_NET:
        #         self.send_gui(
        #             '/ray/gui/client_template_ray_net_update',
        #             int(factory), template_name,
        #             *template_client.ray_net.spread())
        
        # if src_addr is an announced ray GUI
        # server will send it all templates properties
        # else, server replies only templates names
        src_addr_is_gui = False
        server = self.get_server()
        if server is not None:
            src_addr_is_gui = server.is_gui_address(src_addr)

        template_names = set()
        filters = args

        factory = bool('factory' in path)
        base = 'factory' if factory else 'user'

        templates_database = self.get_client_templates_database(base)
        if not templates_database:
            self._rebuild_templates_database(base)
            templates_database = self.get_client_templates_database(base)

        for t in templates_database:
            if filters:
                skipped_by_filter = False
                message = t.template_client.get_properties_message()

                for filt in filters:
                    for line in message.splitlines():
                        if line == filt:
                            break
                    else:
                        skipped_by_filter = True
                        break

                if skipped_by_filter:
                    continue
                
            template_names.add(t.template_name)

        self.send(src_addr, '/reply', path, *template_names)
        
        if src_addr_is_gui:
            for app_template in templates_database:
                template_name = app_template.template_name
                template_client = app_template.template_client
                display_name = app_template.display_name
                
                self.send_gui(
                    '/ray/gui/client_template_update',
                    int(factory), template_name, display_name,
                    *template_client.spread())

                if template_client.protocol == ray.Protocol.RAY_HACK:
                    self.send_gui(
                        '/ray/gui/client_template_ray_hack_update',
                        int(factory), template_name,
                        *template_client.ray_hack.spread())
                elif template_client.protocol == ray.Protocol.RAY_NET:
                    self.send_gui(
                        '/ray/gui/client_template_ray_net_update',
                        int(factory), template_name,
                        *template_client.ray_net.spread())

        self.send(src_addr, '/reply', path)

    def _ray_server_list_factory_client_templates(self, path, args, src_addr):
        self._ray_server_list_client_templates(path, args, src_addr)

    def _ray_server_list_user_client_templates(self, path, args, src_addr):
        self._ray_server_list_client_templates(path, args, src_addr)

    def _ray_server_list_sessions(self, path, args, src_addr):
        with_net = False
        last_sent_time = time.time()

        if args:
            with_net = args[0]

        if with_net:
            for client in self.clients:
                if (client.protocol == ray.Protocol.RAY_NET
                        and client.ray_net.daemon_url):
                    self.send(Address(client.ray_net.daemon_url),
                              '/ray/server/list_sessions', 1)

        if not self.root:
            self.send(src_addr, '/error', path, ray.Err.GENERAL_ERROR,
                      "no session root, so no sessions to list")
            return

        session_list = []
        sessions_set = set()
        n = 0

        for root, dirs, files in os.walk(self.root):
            #exclude hidden files and dirs
            files = [f for f in files if not f.startswith('.')]
            dirs[:] = [d for d in dirs  if not d.startswith('.')]

            if root == self.root:
                continue

            for file in files:
                if file in ('raysession.xml', 'session.nsm'):
                    # prevent search in sub directories
                    dirs.clear()

                    basefolder = root.replace(self.root + '/', '', 1)
                    session_list.append(basefolder)
                    sessions_set.add(basefolder)
                    n += len(basefolder)

                    if n >= 10000 or time.time() - last_sent_time > 0.300:
                        last_sent_time = time.time()
                        self.send(src_addr, "/reply", path, *session_list)

                        session_list.clear()
                        n = 0
                    break

        if session_list:
            self.send(src_addr, "/reply", path, *session_list)

        self.send(src_addr, "/reply", path)
        
        search_scripts_dir = self.root
        has_general_scripts = False
        
        while search_scripts_dir and not search_scripts_dir != '/':
            if os.path.isdir(search_scripts_dir + '/' + ray.SCRIPTS_DIR):
                has_general_scripts = True
                break
            search_scripts_dir = dirname(search_scripts_dir)

        locked_sessions = []
        multi_daemon_file = MultiDaemonFile.get_instance()
        if multi_daemon_file is not None:
            locked_sessions = multi_daemon_file.get_all_session_paths()

        for root, dirs, files in os.walk(self.root):
            #exclude hidden files and dirs
            files = [f for f in files if not f.startswith('.')]
            dirs[:] = [d for d in dirs  if not d.startswith('.')]

            if root == self.root:
                if has_general_scripts:
                    self.send(src_addr, '/ray/gui/listed_session/scripted_dir',
                              '', ray.ScriptFile.PARENT)
                continue
            
            basefolder = root.replace(self.root + '/', '', 1)
            
            if ray.SCRIPTS_DIR in dirs:
                script_files = ray.ScriptFile.PREVENT
                
                for action in ('load', 'save', 'close'):
                    if os.access(
                            "%s/%s/%s.sh" % (root, ray.SCRIPTS_DIR, action),
                            os.X_OK):
                        script_files += ray.ScriptFile.by_string(action)

                self.send(src_addr, '/ray/gui/listed_session/scripted_dir',
                          basefolder, script_files)
            
            if basefolder not in sessions_set:
                continue

            has_notes = bool(ray.NOTES_PATH in files)
            last_modified = int(os.path.getmtime(root))
            locked = bool(root in locked_sessions)

            self.send(src_addr, '/ray/gui/listed_session/details',
                      basefolder, int(has_notes), last_modified, int(locked))

            # prevent search in sub directories
            dirs.clear()        

    def _nsm_server_list(self, path, args, src_addr):
        if self.root:
            for root, dirs, files in os.walk(self.root):
                #exclude hidden files and dirs
                files = [f for f in files if not f.startswith('.')]
                dirs[:] = [d for d in dirs  if not d.startswith('.')]

                if root == self.root:
                    continue

                for file in files:
                    if file in ('raysession.xml', 'session.nsm'):
                        basefolder = root.replace(self.root + '/', '', 1)
                        self.send(src_addr, '/reply', path, basefolder)

        self.send(src_addr, '/reply', path, "")

    @session_operation
    def _ray_server_new_session(self, path, args, src_addr):
        if len(args) == 2 and args[1]:
            session_name, template_name = args

            spath = ''
            if session_name.startswith('/'):
                spath = session_name
            else:
                spath = "%s/%s" % (self.root, session_name)

            if not os.path.exists(spath):
                self.steps_order = [self.save,
                                    self.close_no_save_clients,
                                    self.snapshot,
                                    (self.prepare_template, *args, False),
                                    (self.preload, session_name),
                                    self.close,
                                    self.take_place,
                                    self.load,
                                    self.new_done]
                return

        self.steps_order = [self.save,
                            self.close_no_save_clients,
                            self.snapshot,
                            self.close,
                            (self.new, args[0]),
                            self.save,
                            self.new_done]

    @session_operation
    def _ray_server_open_session(self, path, args, src_addr, open_off=False):
        session_name: str = args[0]
        save_previous = True
        template_name = ''

        if len(args) >= 2:
            save_previous = bool(args[1])
        if len(args) >= 3:
            template_name = args[2]

        if (not session_name
                or '//' in session_name
                or session_name.startswith(('../', '.ray-', 'ray-'))):
            self._send_error(ray.Err.CREATE_FAILED, 'invalid session name.')
            return

        if template_name:
            if '/' in template_name:
                self._send_error(ray.Err.CREATE_FAILED, 'invalid template name')
                return

        spath = ''
        if session_name.startswith('/'):
            spath = session_name
        else:
            spath = "%s/%s" % (self.root, session_name)

        if spath == self.path:
            self._send_error(ray.Err.SESSION_LOCKED,
                _translate('GUIMSG', 'session %s is already opened !')
                    % highlight_text(session_name))
            return

        multi_daemon_file = MultiDaemonFile.get_instance()
        if (multi_daemon_file
                and not multi_daemon_file.is_free_for_session(spath)):
            Terminal.warning("Session %s is used by another daemon"
                              % highlight_text(spath))

            self._send_error(ray.Err.SESSION_LOCKED,
                _translate('GUIMSG',
                    'session %s is already used by another daemon !')
                        % highlight_text(session_name))
            return

        # don't use template if session folder already exists
        if os.path.exists(spath):
            template_name = ''

        self.steps_order = []

        if save_previous:
            self.steps_order += [(self.save, True)]

        self.steps_order += [self.close_no_save_clients]

        if save_previous:
            self.steps_order += [(self.snapshot, '', '', False, True)]

        if template_name:
            self.steps_order += [(self.prepare_template, session_name,
                                 template_name, True)]

        self.steps_order += [(self.preload, session_name),
                             # if open_off, clear all clients at close
                             (self.close, open_off),
                             self.take_place,
                             (self.load, open_off),
                             self.load_done]

    def _ray_server_open_session_off(self, path, args, src_addr):
        self._ray_server_open_session(path, args, src_addr, open_off=True)

    def _ray_server_rename_session(self, path, args, src_addr):
        old_session_name, new_session_name = args
        spath = self.get_full_path(old_session_name)

        for f in 'raysession.xml', 'session.nsm':
            if os.path.isfile("%s/%s" % (spath, f)):
                break
        else:
            self.send(src_addr, '/error', path, ray.Err.BAD_PROJECT,
                      "%s is not an existing session, can't rename !"
                      % old_session_name)

        if '/' in new_session_name:
            self.send(src_addr, '/error', path, ray.Err.BAD_PROJECT,
                      "'/' is not allowed in new_session_name")
            return False

        tmp_session = self._new_dummy_session(self.root)
        tmp_session.ray_server_rename_session(path, args, src_addr)

    def _ray_server_save_session_template(self, path, args, src_addr):
        if len(args) == 2:
            session_name, template_name = args
            sess_root = self.root
            net = False
        else:
            session_name, template_name, sess_root = args
            net = True

        if (sess_root != self.root
                or session_name != self.get_short_path()):
            tmp_session = self._new_dummy_session(sess_root)
            tmp_session.ray_server_save_session_template(
                path, [session_name, template_name, net], src_addr)
            return

        self._ray_session_save_as_template(
            path, [template_name, net], src_addr)

    def _ray_server_get_session_preview(self, path, args, src_addr):
        session_name = args[0]
        server = self.get_server()
        if server is None:
            return
        
        if server.session_to_preview != session_name:
            # prevent to open a dummy session
            # in case user change preview so fastly
            # that this thread is late and user already
            # changed the session to preview
            return

        del self.preview_dummy_session
        self.preview_dummy_session = DummySession(self.root)
        self.preview_dummy_session.ray_server_get_session_preview(
            path, args, src_addr, self._folder_sizes_and_dates)

    def _ray_server_set_option(self, path, args, src_addr):
        option = args[0]

        if abs(option) == ray.Option.BOOKMARK_SESSION:
            if self.path:
                if option > 0:
                    self.bookmarker.make_all(self.path)
                else:
                    self.bookmarker.remove_all(self.path)

    def _ray_server_patchbay_save_group_position(self, path, args, src_addr):
        self.canvas_saver.save_group_position(*args)

    def _ray_server_patchbay_save_portgroup(self, path, args, src_addr):
        self.canvas_saver.save_portgroup(*args)

    @session_operation
    def _ray_session_save(self, path, args, src_addr):        
        self.steps_order = [self.save, self.snapshot, self.save_done]

    @session_operation
    def _ray_session_save_as_template(self, path, args, src_addr):
        template_name = args[0]
        net = False if len(args) < 2 else args[1]

        for client in self.clients:
            if client.protocol == ray.Protocol.RAY_NET:
                client.ray_net.session_template = template_name

        self.steps_order = [self.save, self.snapshot,
                            (self.save_session_template,
                             template_name, net)]

    @session_operation
    def _ray_session_take_snapshot(self, path, args, src_addr):
        if not self.path:
            self.send(src_addr, '/error', path, ray.Err.NO_SESSION_OPEN,
                      'No session is loaded, impossible to take snapshot')
            return
        
        snapshot_name = ''
        with_save = 0
        
        if len(args) == 2:
            snapshot_name, with_save = args
        else:
            snapshot_name = args[0]

        self.steps_order.clear()

        if with_save:
            self.steps_order.append(self.save)
        self.steps_order += [(self.snapshot, snapshot_name, '', True),
                             self.snapshot_done]

    @session_operation
    def _ray_session_close(self, path, args, src_addr):
        self.steps_order = [(self.save, True),
                            self.close_no_save_clients,
                            self.snapshot,
                            (self.close, True),
                            self.close_done]

    def _ray_session_abort(self, path, args, src_addr):
        if not self.path:
            self.file_copier.abort()
            self.send(src_addr, "/error", path, ray.Err.NO_SESSION_OPEN,
                      "No session to abort.")
            return

        self.wait_for = ray.WaitFor.NONE
        self.timer.stop()

        # Non Session Manager can't abort if an operation pending
        # RS can and it would be a big regression to remove this feature
        # So before to abort we need to send an error reply
        # to the last server control message
        # if an operation pending.

        if self.steps_order:
            if self.osc_path.startswith('/nsm/server/'):
                short_path = self.osc_path.rpartition('/')[2]

                if short_path == 'save':
                    self.save_error(ray.Err.CREATE_FAILED)
                elif short_path == 'open':
                    self.load_error(ray.Err.SESSION_LOCKED)
                elif short_path == 'new':
                    self._send_error(ray.Err.CREATE_FAILED,
                                "Could not create the session directory")
                elif short_path == 'duplicate':
                    self.duplicate_aborted(self.osc_args[0])
                elif short_path in ('close', 'abort', 'quit'):
                    # let the current close works here
                    self.send(src_addr, "/error", path,
                              ray.Err.OPERATION_PENDING,
                              "An operation pending.")
                    return
            else:
                self._send_error(ray.Err.ABORT_ORDERED,
                               _translate('GUIMSG',
                                    'abort ordered from elsewhere, sorry !'))

        self.remember_osc_args(path, args, src_addr)
        self.steps_order = [(self.close, True), self.abort_done]

        if self.file_copier.is_active():
            self.file_copier.abort(self.next_function, [])
        else:
            self.next_function()

    def _ray_server_quit(self, path, args, src_addr):
        self.remember_osc_args(path, args, src_addr)
        self.steps_order = [self.terminate_step_scripter,
                            self.close, self.exit_now]

        if self.file_copier.is_active():
            self.file_copier.abort(self.next_function, [])
        else:
            self.next_function()

    def _ray_session_cancel_close(self, path, args, src_addr):
        if not self.steps_order:
            return

        self.timer.stop()
        self.timer_waituser_progress.stop()
        self.steps_order.clear()
        self._clean_expected()
        self.set_server_status(ray.ServerStatus.READY)

    def _ray_session_skip_wait_user(self, path, args, src_addr):
        if not self.steps_order:
            return

        self.timer.stop()
        self.timer_waituser_progress.stop()
        self._clean_expected()
        self.next_function()

    @session_operation
    def _ray_session_duplicate(self, path, args, src_addr):
        new_session_full_name: str = args[0]

        spath = ''
        if new_session_full_name.startswith('/'):
            spath = new_session_full_name
        else:
            spath = "%s/%s" % (self.root, new_session_full_name)

        if os.path.exists(spath):
            self._send_error(ray.Err.CREATE_FAILED,
                _translate('GUIMSG', "%s already exists !")
                    % highlight_text(spath))
            return

        multi_daemon_file = MultiDaemonFile.get_instance()
        if (multi_daemon_file
                and not multi_daemon_file.is_free_for_session(spath)):
            Terminal.warning("Session %s is used by another daemon"
                             % highlight_text(new_session_full_name))
            self._send_error(ray.Err.SESSION_LOCKED,
                _translate('GUIMSG',
                    'session %s is already used by this or another daemon !')
                        % highlight_text(new_session_full_name))
            return

        self.steps_order = [self.save,
                            self.close_no_save_clients,
                            self.snapshot,
                            (self.duplicate, new_session_full_name),
                            (self.preload, new_session_full_name),
                            self.close,
                            self.take_place,
                            self.load,
                            self.duplicate_done]

    def _ray_session_duplicate_only(self, path, args: list[str], src_addr):
        session_to_load, new_session, sess_root = args

        spath = ''
        if new_session.startswith('/'):
            spath = new_session
        else:
            spath = "%s/%s" % (sess_root, new_session)

        if os.path.exists(spath):
            self.send(src_addr, '/ray/net_daemon/duplicate_state', 1)
            self.send(src_addr, '/error', path, ray.Err.CREATE_FAILED,
                      _translate('GUIMSG', "%s already exists !")
                        % highlight_text(spath))
            return

        if sess_root == self.root and session_to_load == self.get_short_path():
            if (self.steps_order
                    or self.file_copier.is_active()):
                self.send(src_addr, '/ray/net_daemon/duplicate_state', 1)
                return

            self.remember_osc_args(path, args, src_addr)

            self.steps_order = [self.save,
                                self.snapshot,
                                (self.duplicate, new_session),
                                self.duplicate_only_done]

            self.next_function()

        else:
            tmp_session = self._new_dummy_session(sess_root)
            tmp_session.osc_src_addr = src_addr
            tmp_session.dummy_duplicate(path, args, src_addr)

    @session_operation
    def _ray_session_open_snapshot(self, path, args, src_addr):
        if not self.path:
            return

        snapshot = args[0]

        self.steps_order = [self.save,
                              self.close_no_save_clients,
                              (self.snapshot, '', snapshot, True),
                              (self.close, True),
                              (self.init_snapshot, self.path, snapshot),
                              (self.preload, self.path),
                              self.take_place,
                              self.load,
                              self.load_done]

    def _ray_session_rename(self, path, args, src_addr):
        new_session_name = args[0]

        if self.steps_order:
            return

        if not self.path:
            return

        if self.file_copier.is_active():
            return

        if new_session_name == self.name:
            return

        if not self.is_nsm_locked():
            for filename in os.listdir(dirname(self.path)):
                if filename == new_session_name:
                    # another directory exists with new session name
                    return

        for client in self.clients:
            if client.is_running():
                self.send_gui_message(
                    _translate('GUIMSG',
                               'Stop all clients before rename session !'))
                return

        for client in self.clients + self.trashed_clients:
            client.adjust_files_after_copy(new_session_name, ray.Template.RENAME)

        if not self.is_nsm_locked():
            try:
                spath = "%s/%s" % (dirname(self.path), new_session_name)
                subprocess.run(['mv', self.path, spath])
                self._set_path(spath)

                self.send_gui_message(
                    _translate('GUIMSG', 'Session directory is now: %s')
                    % self.path)
            except:
                pass

        # we need to save the session file here
        # because session just has been renamed
        # and clients dependant of the session name
        # would not find there files if session is aborted just after
        self._save_session_file()

        self.send_gui_message(
            _translate('GUIMSG', 'Session %s has been renamed to %s .')
            % (self.name, new_session_name))
        self.send_gui('/ray/gui/session/name', self.name, self.path)

    def _ray_session_set_notes(self, path, args, src_addr):
        self.notes = args[0]
        self.send(src_addr, '/reply', path, 'Notes has been set')

    def _ray_session_get_notes(self, path, args, src_addr):
        self.send(src_addr, '/reply', path, self.notes)
        self.send(src_addr, '/reply', path)

    def _ray_session_add_executable(self, path: str, args, src_addr):
        protocol = ray.Protocol.NSM
        executable = args[0]
        via_proxy = 0
        prefix_mode = ray.PrefixMode.SESSION_NAME
        custom_prefix = ''
        client_id = ""
        start_it = 1
        jack_naming = 0

        if len(args) == 1:
            pass

        elif ray.are_they_all_strings(args):
            if TYPE_CHECKING:
                assert isinstance(args, list[str])
            
            via_proxy = int(bool('via_proxy' in args[1:]))
            start_it = int(bool('not_start' not in args[1:]))
            if 'ray_hack' in args[1:]:
                protocol = ray.Protocol.RAY_HACK

            for arg in args[1:]:
                if arg == 'prefix_mode:client_name':
                    prefix_mode = ray.PrefixMode.CLIENT_NAME

                elif arg == 'prefix_mode:session_name':
                    prefix_mode = ray.PrefixMode.SESSION_NAME

                elif arg.startswith('prefix:'):
                    custom_prefix = arg.partition(':')[2]
                    if not custom_prefix or '/' in custom_prefix:
                        self.send(src_addr, '/error', path,
                                  ray.Err.CREATE_FAILED,
                                  "wrong custom prefix !")
                        return

                    prefix_mode = ray.PrefixMode.CUSTOM

                elif arg.startswith('client_id:'):
                    client_id = arg.partition(':')[2]
                    if not client_id.replace('_', '').isalnum():
                        self.send(src_addr, '/error', path,
                                  ray.Err.CREATE_FAILED,
                                  f"client_id {client_id} is not alphanumeric")
                        return

                    # Check if client_id already exists
                    for client in self.clients + self.trashed_clients:
                        if client.client_id == client_id:
                            self.send(src_addr, '/error', path,
                                ray.Err.CREATE_FAILED,
                                "client_id %s is already used" % client_id)
                            return

                elif arg.startswith('jack_naming:'):
                    str_jack_naming = arg.partition(':')[2]
                    if str_jack_naming.lower() in ('1', 'long'):
                        jack_naming = 1

        else:
            executable, start_it, protocol, \
                prefix_mode, custom_prefix, client_id, jack_naming = args

            if prefix_mode == ray.PrefixMode.CUSTOM and not custom_prefix:
                prefix_mode = ray.PrefixMode.SESSION_NAME

            client_id: str

            if client_id:
                if not client_id.replace('_', '').isalnum():
                    self.send(src_addr, '/error', path, ray.Err.CREATE_FAILED,
                      _translate("error", "client_id %s is not alphanumeric")
                        % client_id)
                    return

                # Check if client_id already exists
                for client in self.clients + self.trashed_clients:
                    if client.client_id == client_id:
                        self.send(src_addr, '/error', path,
                          ray.Err.CREATE_FAILED,
                          _translate("error", "client_id %s is already used")
                            % client_id)
                        return

        if not client_id:
            client_id = self.generate_client_id(executable)

        client = Client(self)

        client.protocol = protocol
        if client.protocol == ray.Protocol.NSM and via_proxy:
            client.executable_path = 'ray-proxy'
        else:
            client.executable_path = executable
        client.name = os.path.basename(executable)
        client.client_id = client_id
        client.prefix_mode = prefix_mode
        client.custom_prefix = custom_prefix
        client.set_default_git_ignored(executable)
        client.jack_naming = jack_naming

        if self._add_client(client):
            if start_it:
                client.start()

            reply_str = client.client_id
            if path.startswith('/nsm/server/'):
                reply_str = "Launched."

            self.send(src_addr, '/reply', path, reply_str)
        else:
            self.send(src_addr, '/error', path, ray.Err.NOT_NOW,
                      "Impossible to add client now")

    def _ray_session_add_client_template(self, path, args, src_addr):
        if not self.path:
            self.send(src_addr, "/error", path, ray.Err.NO_SESSION_OPEN,
                      "Cannot add to session because no session is loaded.")
            return

        factory = bool(args[0])
        template_name: str = args[1]
        auto_start = bool(len(args) <= 2 or args[2] != 'not_start')
        unique_id: str = args[3] if len(args) > 3 else ''

        if unique_id:
            if not unique_id.replace('_', '').isalnum():
                self.send(src_addr, '/error', path,
                            ray.Err.CREATE_FAILED,
                            f"client_id {unique_id} is not alphanumeric")
                return

            # Check if client_id already exists
            for client in self.clients + self.trashed_clients:
                if client.client_id == unique_id:
                    self.send(src_addr, '/error', path,
                        ray.Err.CREATE_FAILED,
                        f"client_id {unique_id} is already used")
                    return

        self.add_client_template(
            src_addr, path, template_name, factory, auto_start, unique_id)

    def _ray_session_add_factory_client_template(self, path, args, src_addr):
        self._ray_session_add_client_template(path, [1] + args, src_addr)

    def _ray_session_add_user_client_template(self, path, args, src_addr):
        self._ray_session_add_client_template(path, [0] + args, src_addr)

    @session_operation
    def _ray_session_add_other_session_client(self, path, args, src_addr):
        other_session, client_id = args    

        # @session_operation remember them but this is not needed here
        self._forget_osc_args()

        dummy_session = DummySession(self.root)
        dummy_session.dummy_load(other_session)
        
        # hopefully for a dummy session,
        # there is nothing to wait to have a loaded session
        # This is quite dirty but so easier

        if not dummy_session.path:
            self.send(src_addr, '/error', path, ray.Err.NOT_NOW,
                      "falied to load other session %s" % other_session)
            return
        
        for client in dummy_session.clients:
            if client.client_id == client_id:
                new_client = Client(self)
                new_client.client_id = self.generate_client_id(
                    Client.short_client_id(client_id))

                ok = self._add_client(new_client)
                if not ok:
                    self.send(src_addr, '/error', path, ray.Err.NOT_NOW,
                              'session is busy')
                    return

                new_client.eat_other_session_client(src_addr, path, client)
                break
        else:
            self.send(src_addr, '/error', path, ray.Err.NO_SUCH_FILE,
                      'no client %s found in session %s' % (client_id, other_session))

    def _ray_session_reorder_clients(self, path, args, src_addr):
        client_ids_list = args

        if not self.path:
            self.send(src_addr, '/error', path, ray.Err.NO_SESSION_OPEN,
                      "no session to reorder clients")

        if len(self.clients) < 2:
            self.send(src_addr, '/reply', path, "clients reordered")
            return

        self._re_order_clients(client_ids_list, src_addr, path)

    def _ray_session_clear_clients(self, path, args, src_addr):
        if not self.load_locked:
            self.send(src_addr, '/error', path, ray.Err.NOT_NOW,
                "clear_clients has to be used only during the load script !")
            return

        self.clear_clients(src_addr, path, *args)

    def _ray_session_list_snapshots(self, path, args, src_addr, client_id=""):
        if not self.path:
            self.send(src_addr, '/error', path, ray.Err.NO_SESSION_OPEN,
                      "no session to list snapshots")
            return

        auto_snapshot = not self.snapshoter.is_auto_snapshot_prevented()
        self.send_gui('/ray/gui/session/auto_snapshot', int(auto_snapshot))

        snapshots = self.snapshoter.list(client_id)

        i = 0
        snap_send = []

        for snapshot in snapshots:
            if i == 20:
                self.send(src_addr, '/reply', path, *snap_send)

                snap_send.clear()
                i = 0
            else:
                snap_send.append(snapshot)
                i += 1

        if snap_send:
            self.send(src_addr, '/reply', path, *snap_send)
        self.send(src_addr, '/reply', path)

    def _ray_session_set_auto_snapshot(self, path, args, src_addr):
        self.snapshoter.set_auto_snapshot(bool(args[0]))

    def _ray_session_list_clients(self, path, args: list[str], src_addr):
        if not self.path:
            self.send(src_addr, '/error', path, ray.Err.NO_SESSION_OPEN,
                      _translate('GUIMSG', 'No session to list clients !'))
            return

        f_started = -1
        f_active = -1
        f_auto_start = -1
        f_no_save_level = -1

        search_properties = list[tuple[int, str]]()

        for arg in args:
            cape = 1
            if arg.startswith('not_'):
                cape = 0
                arg = arg.replace('not_', '', 1)

            if ':' in arg:
                search_properties.append((cape, arg))

            elif arg == 'started':
                f_started = cape
            elif arg == 'active':
                f_active = cape
            elif arg == 'auto_start':
                f_auto_start = cape
            elif arg == 'no_save_level':
                f_no_save_level = cape

        client_id_list = list[str]()

        for client in self.clients:
            if ((f_started < 0 or f_started == client.is_running())
                and (f_active < 0 or f_active == client.nsm_active)
                and (f_auto_start < 0 or f_auto_start == client.auto_start)
                and (f_no_save_level < 0
                     or f_no_save_level == int(bool(client.noSaveLevel())))):
                if search_properties:
                    message = client.get_properties_message()

                    for cape, search_prop in search_properties:
                        line_found = False

                        for line in message.split('\n'):
                            if line == search_prop:
                                line_found = True
                                break

                        if cape != line_found:
                            break
                    else:
                        client_id_list.append(client.client_id)
                else:
                    client_id_list.append(client.client_id)

        if client_id_list:
            self.send(src_addr, '/reply', path, *client_id_list)
        self.send(src_addr, '/reply', path)

    def _ray_session_list_trashed_clients(self, path, args, src_addr):
        client_id_list = list[str]()

        for trashed_client in self.trashed_clients:
            client_id_list.append(trashed_client.client_id)

        if client_id_list:
            self.send(src_addr, '/reply', path, *client_id_list)
        self.send(src_addr, '/reply', path)

    def _ray_session_run_step(self, path, args, src_addr):
        if not self.step_scripter.is_running():
            self.send(src_addr, '/error', path, ray.Err.GENERAL_ERROR,
              'No stepper script running, run run_step from session scripts')
            return

        if self.step_scripter.stepper_has_called():
            self.send(src_addr, '/error', path, ray.Err.GENERAL_ERROR,
             'step already done. Run run_step only one time in the script')
            return

        if not self.steps_order:
            self.send(src_addr, '/error', path, ray.Err.GENERAL_ERROR,
                      'No operation pending !')
            return

        self.run_step_addr = src_addr
        self.next_function(True, args)

    @client_action
    def _ray_client_stop(self, path, args, src_addr, client:Client):
        client.stop(src_addr, path)

    @client_action
    def _ray_client_kill(self, path, args, src_addr, client:Client):
        client.kill()
        self.send(src_addr, "/reply", path, "Client killed.")

    @client_action
    def _ray_client_trash(self, path, args, src_addr, client:Client):
        if client.is_running():
            self.send(src_addr, '/error', path, ray.Err.OPERATION_PENDING,
                        "Stop client before to trash it !")
            return

        if self.file_copier.is_active(client.client_id):
            self.file_copier.abort()
            self.send(src_addr, '/error', path, ray.Err.COPY_RUNNING,
                        "Files were copying for this client.")
            return

        self._trash_client(client)

        self.send(src_addr, "/reply", path, "Client removed.")

    def _ray_client_start(self, path, args, src_addr):
        self._ray_client_resume(path, args, src_addr)

    @client_action
    def _ray_client_resume(self, path, args, src_addr, client:Client):
        if client.is_running():
            self.send_gui_message(
                _translate('GUIMSG', 'client %s is already running.')
                    % client.gui_msg_style())

            # make ray_control exit code 0 in this case
            self.send(src_addr, '/reply', path, 'client running')
            return

        if self.file_copier.is_active(client.client_id):
            self.send_error_copy_running(src_addr, path)
            return

        client.start(src_addr, path)

    @client_action
    def _ray_client_open(self, path, args, src_addr, client:Client):
        if self.file_copier.is_active(client.client_id):
            self.send_error_copy_running(src_addr, path)
            return

        if client.nsm_active:
            self.send_gui_message(
                _translate('GUIMSG', 'client %s is already active.')
                    % client.gui_msg_style())

            # make ray_control exit code 0 in this case
            self.send(src_addr, '/reply', path, 'client active')
        else:
            client.load(src_addr, path)

    @client_action
    def _ray_client_save(self, path, args, src_addr, client:Client):
        if client.can_save_now():
            if self.file_copier.is_active(client.client_id):
                self.send_error_copy_running(src_addr, path)
                return
            client.save(src_addr, path)
        else:
            self.send_gui_message(_translate('GUIMSG', "%s is not saveable.")
                                    % client.gui_msg_style())
            self.send(src_addr, '/reply', path, 'client saved')

    @client_action
    def _ray_client_save_as_template(self, path, args, src_addr, client:Client):
        template_name = args[0]

        if self.file_copier.is_active():
            self.send_error_copy_running(src_addr, path)
            return

        client.save_as_template(template_name, src_addr, path)

    @client_action
    def _ray_client_show_optional_gui(self, path, args, src_addr, client:Client):
        client.send_to_self_address("/nsm/client/show_optional_gui")
        client.show_gui_ordered = True
        self.send(src_addr, '/reply', path, 'show optional GUI asked')

    @client_action
    def _ray_client_hide_optional_gui(self, path, args, src_addr, client:Client):
        client.send_to_self_address("/nsm/client/hide_optional_gui")
        self.send(src_addr, '/reply', path, 'hide optional GUI asked')

    @client_action
    def _ray_client_update_properties(self, path, args, src_addr, client:Client):
        client.update_secure(client.client_id, *args)
        client.send_gui_client_properties()
        self.send(src_addr, '/reply', path, 'client properties updated')

    @client_action
    def _ray_client_update_ray_hack_properties(self, path, args,
                                               src_addr, client:Client):
        ex_no_save_level = client.noSaveLevel()

        if client.is_ray_hack():
            client.ray_hack.update(*args)

        no_save_level = client.noSaveLevel()

        if no_save_level != ex_no_save_level:
            self.send_gui('/ray/gui/client/no_save_level',
                           client.client_id, no_save_level)

        self.send(src_addr, '/reply', path, 'ray_hack updated')

    @client_action
    def _ray_client_update_ray_net_properties(self, path, args,
                                              src_addr, client:Client):
        if client.protocol == ray.Protocol.RAY_NET:
            client.ray_net.update(*args)
        self.send(src_addr, '/reply', path, 'ray_net updated')

    @client_action
    def _ray_client_set_properties(self, path, args, src_addr, client:Client):
        message = ''
        for arg in args:
            message += "%s\n" % arg

        client.set_properties_from_message(message)
        self.send(src_addr, '/reply', path,
                    'client properties updated')

    @client_action
    def _ray_client_get_properties(self, path, args, src_addr, client:Client):
        message = client.get_properties_message()
        self.send(src_addr, '/reply', path, message)
        self.send(src_addr, '/reply', path)

    @client_action
    def _ray_client_get_proxy_properties(self, path, args, src_addr,
                                         client:Client):
        proxy_file = '%s/ray-proxy.xml' % client.get_project_path()

        if not os.path.isfile(proxy_file):
            self.send(src_addr, '/error', path, ray.Err.GENERAL_ERROR,
                _translate('GUIMSG', '%s seems to not be a proxy client !')
                    % client.gui_msg_style())
            return

        try:
            file = open(proxy_file, 'r')
            xml = QDomDocument()
            xml.setContent(file.read())
            content = xml.documentElement()
            file.close()
        except:
            self.send(src_addr, '/error', path, ray.Err.BAD_PROJECT,
                _translate('GUIMSG', "impossible to read %s correctly !")
                    % proxy_file)
            return

        if content.tagName() != "RAY-PROXY":
            self.send(src_addr, '/error', path, ray.Err.BAD_PROJECT,
                _translate('GUIMSG', "impossible to read %s correctly !")
                    % proxy_file)
            return

        cte = content.toElement()
        message = ""
        for prop in ('executable', 'arguments', 'config_file',
                     'save_signal', 'stop_signal',
                     'no_save_level', 'wait_window',
                     'VERSION'):
            message += "%s:%s\n" % (prop, cte.attribute(prop))

        # remove last empty line
        message = message.rpartition('\n')[0]

        self.send(src_addr, '/reply', path, message)
        self.send(src_addr, '/reply', path)

    @client_action
    def _ray_client_set_proxy_properties(self, path, args, src_addr,
                                         client:Client):
        message = ''
        for arg in args:
            message += "%s\n" % arg

        if client.is_running():
            self.send(src_addr, '/error', path, ray.Err.GENERAL_ERROR,
              _translate('GUIMSG',
               'Impossible to set proxy properties while client is running.'))
            return

        proxy_file = '%s/ray-proxy.xml' % client.get_project_path()

        if (not os.path.isfile(proxy_file)
                and client.executable_path != 'ray-proxy'):
            self.send(src_addr, '/error', path, ray.Err.GENERAL_ERROR,
                _translate('GUIMSG', '%s seems to not be a proxy client !')
                    % client.gui_msg_style())
            return

        if os.path.isfile(proxy_file):
            try:
                file = open(proxy_file, 'r')
                xml = QDomDocument()
                xml.setContent(file.read())
                content = xml.documentElement()
                file.close()
            except:
                self.send(src_addr, '/error', path, ray.Err.BAD_PROJECT,
                    _translate('GUIMSG', "impossible to read %s correctly !")
                        % proxy_file)
                return
        else:
            xml = QDomDocument()
            p = xml.createElement('RAY-PROXY')
            p.setAttribute('VERSION', ray.VERSION)
            xml.appendChild(p)
            content = xml.documentElement()

            if not os.path.isdir(client.get_project_path()):
                try:
                    os.makedirs(client.get_project_path())
                except:
                    self.send(src_addr, '/error', path, ray.Err.CREATE_FAILED,
                              "Impossible to create proxy directory")
                    return

        if content.tagName() != "RAY-PROXY":
            self.send(src_addr, '/error', path, ray.Err.BAD_PROJECT,
                _translate('GUIMSG', "impossible to read %s correctly !")
                    % proxy_file)
            return

        cte = content.toElement()

        for line in message.split('\n'):
            prop, colon, value = line.partition(':')
            if prop in (
                    'executable', 'arguments',
                    'config_file', 'save_signal', 'stop_signal',
                    'no_save_level', 'wait_window', 'VERSION'):
                cte.setAttribute(prop, value)

        try:
            file = open(proxy_file, 'w')
            file.write(xml.toString())
            file.close()
        except:
            self.send(src_addr, '/error', path, ray.Err.BAD_PROJECT,
                _translate('GUIMSG', "%s is not writeable")
                    % proxy_file)
            return

        self.send(src_addr, '/reply', path, message)
        self.send(src_addr, '/reply', path)

    @client_action
    def _ray_client_get_description(self, path, args, src_addr, client:Client):
        self.send(src_addr, '/reply', path, client.description)
        self.send(src_addr, '/reply', path)

    @client_action
    def _ray_client_set_description(self, path, args, src_addr, client:Client):
        client.description = args[0]
        self.send(src_addr, '/reply', path, 'Description updated')

    @client_action
    def _ray_client_list_files(self, path, args, src_addr, client:Client):
        client_files = client.get_project_files()
        self.send(src_addr, '/reply', path, *client_files)
        self.send(src_addr, '/reply', path)

    @client_action
    def _ray_client_get_pid(self, path, args, src_addr, client:Client):
        if client.is_running():
            self.send(src_addr, '/reply', path, str(client.pid))
            self.send(src_addr, '/reply', path)
        else:
            self.send(src_addr, '/error', path, ray.Err.NOT_NOW,
                "client is not running, impossible to get its pid")

    def _ray_client_list_snapshots(self, path, args, src_addr):
        self._ray_session_list_snapshots(path, [], src_addr, args[0])

    @session_operation
    def _ray_client_open_snapshot(self, path, args, src_addr):
        client_id, snapshot = args

        for client in self.clients:
            if client.client_id == client_id:
                if client.is_running():
                    self.steps_order = [
                        self.save,
                        (self.snapshot, '', snapshot, True),
                        self.before_close_client_for_snapshot,
                        (self.close_client, client),
                        (self.load_client_snapshot, client_id, snapshot),
                        (self.start_client, client),
                        self.load_client_snapshot_done]
                else:
                    self.steps_order = [
                        self.save,
                        (self.snapshot, '', snapshot, True),
                        (self.load_client_snapshot, client_id, snapshot),
                        self.load_client_snapshot_done]
                break
        else:
            self.send_error_no_client(src_addr, path, client_id)

    @client_action
    def _ray_client_is_started(self, path, args, src_addr, client:Client):
        if client.is_running():
            self.send(src_addr, '/reply', path, 'client running')
        else:
            self.send(src_addr, '/error', path, ray.Err.GENERAL_ERROR,
                        _translate('GUIMSG', '%s is not running.')
                        % client.gui_msg_style())

    @client_action
    def _ray_client_send_signal(self, path, args, src_addr, client:Client):
        sig = args[0]
        client.send_signal(sig, src_addr, path)

    @client_action
    def _ray_client_set_custom_data(self, path, args, src_addr, client:Client):
        data, value = args
        client.custom_data[data] = value
        self.send(src_addr, '/reply', path, 'custom data set')

    @client_action
    def _ray_client_get_custom_data(self, path, args, src_addr, client:Client):
        data = args[0]

        if data not in client.custom_data:
            self.send(src_addr, '/error', path, ray.Err.NO_SUCH_FILE,
                        "client %s has no custom_data key '%s'"
                        % (client.client_id, data))
            return

        self.send(src_addr, '/reply', path, client.custom_data[data])
        self.send(src_addr, '/reply', path)

    @client_action
    def _ray_client_set_tmp_data(self, path, args, src_addr, client:Client):
        data, value = args
        client.custom_tmp_data[data] = value
        self.send(src_addr, '/reply', path, 'custom tmp data set')

    @client_action
    def _ray_client_get_tmp_data(self, path, args, src_addr, client:Client):
        data = args[0]

        if data not in client.custom_tmp_data:
            self.send(src_addr, '/error', path, ray.Err.NO_SUCH_FILE,
                      "client %s has no tmp_custom_data key '%s'"
                        % (client.client_id, data))
            return

        self.send(src_addr, '/reply', path, client.custom_tmp_data[data])
        self.send(src_addr, '/reply', path)

    @client_action
    def _ray_client_change_prefix(self, path, args, src_addr, client:Client):
        if client.is_running():
            self.send(src_addr, '/error', path, ray.Err.NOT_NOW,
                      "impossible to change prefix while client is running")
            return

        prefix_mode = args[0]
        custom_prefix = ''

        if prefix_mode in (ray.PrefixMode.SESSION_NAME, 'session_name'):
            prefix_mode = ray.PrefixMode.SESSION_NAME
        elif prefix_mode in (ray.PrefixMode.CLIENT_NAME, 'client_name'):
            prefix_mode = ray.PrefixMode.CLIENT_NAME
        else:
            prefix_mode = ray.PrefixMode.CUSTOM

        if prefix_mode == ray.PrefixMode.CUSTOM:
            custom_prefix = args[1]
            if not custom_prefix:
                self.send(
                    src_addr, '/error', path, ray.Err.GENERAL_ERROR,
                    "You need to specify a custom prefix as 2nd argument")
                return

        client.change_prefix(prefix_mode, custom_prefix)
        
        # we need to save session file here
        # else, if session is aborted
        # client won't find its files at next restart
        self._save_session_file()

        self.send(src_addr, '/reply', path, 'prefix changed')

    @client_action
    def _ray_client_change_advanced_properties(self, path, args, src_addr, client: Client):
        if client.is_running():
            self.send(src_addr, '/error', path, ray.Err.NOT_NOW,
                      "impossible to change id while client is running")
            return

        new_client_id: str
        prefix_mode: int
        custom_prefix: str
        jack_naming: int

        new_client_id, prefix_mode, custom_prefix, jack_naming = args

        if new_client_id != client.client_id:
            if new_client_id in [c.client_id for c in
                                 self.clients + self.trashed_clients]:
                self.send(src_addr, '/error', path, ray.Err.BLACKLISTED,
                        f"client id '{new_client_id}' already exists in the session")
                return

        if prefix_mode == ray.PrefixMode.CUSTOM and not custom_prefix:
            self.send(src_addr, '/error', path, ray.Err.BAD_PROJECT,
                      "Custom prefix missing is missing !")
            return

        tmp_client = Client(self)
        tmp_client.eat_attributes(client)
        tmp_client.client_id = new_client_id
        tmp_client.prefix_mode = prefix_mode
        tmp_client.custom_prefix = custom_prefix
        tmp_client.jack_naming = jack_naming
        
        client.set_status(ray.ClientStatus.REMOVED)
        
        client._rename_files(
            Path(self.path),
            self.name, self.name,
            client.get_prefix_string(), tmp_client.get_prefix_string(),
            client.client_id, tmp_client.client_id,
            client.get_links_dir(), tmp_client.get_links_dir())

        ex_jack_name = client.get_jack_client_name()
        ex_client_id = client.client_id
        new_jack_name = tmp_client.get_jack_client_name()

        client.client_id = new_client_id
        client.prefix_mode = prefix_mode
        client.custom_prefix = custom_prefix
        client.jack_naming = jack_naming
        self._update_forbidden_ids_set()

        if new_jack_name != ex_jack_name:
            rewrite_jack_patch_files(
                self, ex_client_id, new_client_id,
                ex_jack_name, new_jack_name)
            self.canvas_saver.client_jack_name_changed(
                ex_jack_name, new_jack_name)

        client.sent_to_gui = False
        client.send_gui_client_properties()
        self.send_gui('/ray/gui/session/sort_clients',
                      *[c.client_id for c in self.clients])

        # we need to save session file here
        # else, if session is aborted
        # client won't find its files at next restart
        self._save_session_file()

        self.send_monitor_event('id_changed_to:' + new_client_id, ex_client_id)
        self.send(src_addr, '/reply', path, 'client id changed')

    @client_action
    def _ray_client_full_rename(self, path, args, src_addr, client: Client):
        if self.steps_order:
            self.send(src_addr, '/error', path, ray.Err.NOT_NOW,
                      "Session is not ready for full client rename")
            return
        
        new_client_name: str = args[0]
        new_client_id = new_client_name.replace(' ', '_')

        if not new_client_id or not new_client_id.replace('_', '').isalnum():
            self.send(src_addr, '/error', path, ray.Err.BAD_PROJECT,
                      f'client_id {new_client_id} is not alphanumeric')
            return

        if new_client_id in self.forbidden_ids_set:
            self.send(src_addr, '/error', path, ray.Err.BAD_PROJECT,
                      f'client_id {new_client_id} is forbidden in this session')
            return

        if client.is_running():
            if not client.status == ray.ClientStatus.READY:
                self.send(src_addr, '/error', path, ray.Err.NOT_NOW,
                          f'client_id {new_client_id} is not ready')
                return

            elif client.is_capable_of(':switch:'):
                self.steps_order = [
                    (self.save_client_and_patchers, client),
                    (self.rename_full_client, client, new_client_name, new_client_id),
                    (self.switch_client, client),
                    (self.rename_full_client_done, client)]
            else:
                self.steps_order = [
                    (self.save_client_and_patchers, client),
                    (self.close_client, client),
                    (self.rename_full_client, client, new_client_name, new_client_id),
                    (self.restart_client, client),
                    (self.rename_full_client_done, client)
                ]
        else:
            self.steps_order = [
                (self.rename_full_client, client, new_client_name, new_client_id),
                (self.rename_full_client_done, client)
            ]

        self.next_function()

    @client_action
    def _ray_client_change_id(self, path, args, src_addr, client: Client):
        if client.is_running():
            self.send(src_addr, '/error', path, ray.Err.NOT_NOW,
                      "impossible to change id while client is running")
            return

        new_client_id: str = args[0]
        
        if new_client_id in [c.client_id for c in
                             self.clients + self.trashed_clients]:
            self.send(src_addr, '/error', path, ray.Err.BLACKLISTED,
                      f"client id '{new_client_id}' already exists in the session")
            return
        
        if not new_client_id.replace('_', '').isalnum():
            self.send(src_addr, '/error', path, ray.Err.BAD_PROJECT,
                      f"client id {new_client_id} contains forbidden characters")
            return
        
        ex_client_id = client.client_id
        ex_jack_name = client.get_jack_client_name()
        client.set_status(ray.ClientStatus.REMOVED)

        prefix = client.get_prefix_string()
        links_dir = client.get_links_dir()

        client._rename_files(
            Path(self.path),
            self.name, self.name,
            prefix, prefix,
            ex_client_id, new_client_id,
            links_dir, links_dir)

        client.client_id = new_client_id
        self._update_forbidden_ids_set()
        new_jack_name = client.get_jack_client_name()

        if new_jack_name != ex_jack_name:
            rewrite_jack_patch_files(
                self, ex_client_id, new_client_id,
                ex_jack_name, new_jack_name)
            self.canvas_saver.client_jack_name_changed(
                ex_jack_name, new_jack_name)

        client.sent_to_gui = False
        client.send_gui_client_properties()
        self.send_gui('/ray/gui/session/sort_clients',
                      *[c.client_id for c in self.clients])

        # we need to save session file here
        # else, if session is aborted
        # client won't find its files at next restart
        self._save_session_file()

        self.send_monitor_event('id_changed_to:' + new_client_id, ex_client_id)
        self.send(src_addr, '/reply', path, 'client id changed')
    
    def _ray_trashed_client_restore(self, path, args, src_addr):
        if not self.path:
            self.send(src_addr, "/error", path, ray.Err.NO_SESSION_OPEN,
                      "Nothing in trash because no session is loaded.")
            return

        for client in self.trashed_clients:
            if client.client_id == args[0]:
                if self._restore_client(client):
                    self.send(src_addr, '/reply', path, "client restored")
                else:
                    self.send(src_addr, '/error', path, ray.Err.NOT_NOW,
                              "Session is in a loading locked state")
                break
        else:
            self.send(src_addr, "/error", path, -10, "No such client.")

    def _ray_trashed_client_remove_definitely(self, path, args, src_addr):
        if not self.path:
            self.send(src_addr, "/error", path, ray.Err.NO_SESSION_OPEN,
                      "Nothing in trash because no session is loaded.")
            return

        client_id = args[0]

        for client in self.trashed_clients:
            if client.client_id == client_id:
                break
        else:
            self.send(src_addr, "/error", path, -10, "No such client.")
            return

        self.send_gui('/ray/gui/trash/remove', client.client_id)

        for file in client.get_project_files():
            try:
                subprocess.run(['rm', '-R', file])
            except:
                self.send(src_addr, '/minor_error', path, -10,
                          "Error while removing client file %s" % file)
                continue

        self.trashed_clients.remove(client)
        self._save_session_file()

        self.send(src_addr, '/reply', path, "client definitely removed")
        self.send_monitor_event('removed', client_id)

    def _ray_trashed_client_remove_keep_files(self, path, args, src_addr):
        if not self.path:
            self.send(src_addr, "/error", path, ray.Err.NO_SESSION_OPEN,
                      "Nothing in trash because no session is loaded.")
            return

        client_id = args[0]

        for client in self.trashed_clients:
            if client.client_id == client_id:
                break
        else:
            self.send(src_addr, "/error", path, -10, "No such client.")
            return

        self.send_gui('/ray/gui/trash/remove', client.client_id)

        self.trashed_clients.remove(client)

        self.send(src_addr, '/reply', path, "client removed")
        self.send_monitor_event('removed', client_id)

    def _ray_net_daemon_duplicate_state(self, path, args, src_addr):
        state = args[0]
        for client in self.clients:
            if (client.protocol == ray.Protocol.RAY_NET
                    and client.ray_net.daemon_url
                    and ray.are_same_osc_port(client.ray_net.daemon_url,
                                              src_addr.url)):
                client.ray_net.duplicate_state = state
                client.net_daemon_copy_timer.stop()
                break
        else:
            return

        if state == 1:
            if self.wait_for == ray.WaitFor.DUPLICATE_FINISH:
                self.end_timer_if_last_expected(client)
            return

        if (self.wait_for == ray.WaitFor.DUPLICATE_START and state == 0):
            self.end_timer_if_last_expected(client)

        client.net_daemon_copy_timer.start()

    def check_recent_sessions_existing(self):
        # check here if recent sessions still exist
        if self.root in self.recent_sessions.keys():
            to_remove_list = list[str]()
            for sess in self.recent_sessions[self.root]:
                if not os.path.exists(
                        "%s/%s/raysession.xml" % (self.root, sess)):
                    to_remove_list.append(sess)
            for sess in to_remove_list:
                self.recent_sessions[self.root].remove(sess)

    def server_open_session_at_start(self, session_name):
        self.steps_order = [(self.preload, session_name),
                            self.take_place,
                            self.load,
                            self.load_done]
        self.next_function()

    def dummy_load_and_template(self, session_name, template_name, sess_root):
        tmp_session = self._new_dummy_session(sess_root)
        tmp_session.dummy_load_and_template(session_name, template_name)

    def terminate(self):
        if self.terminated_yet:
            return

        if self.file_copier.is_active():
            self.file_copier.abort()

        self.terminated_yet = True
        self.steps_order = [self.terminate_step_scripter,
                            self.close, self.exit_now]
        self.next_function()


class DummySession(OperatingSession):
    ''' A dummy session allows to make such operations on not current session
        It is used for session preview, or duplicate a session for example.
        When a session is dummy, it has no server options
        (bookmarks, snapshots, session scripts...).
        All clients are dummy and can't be started.
        Their file copier is not dummy, it can send OSC messages to gui,
        That is why we need a session_id to find it '''

    def __init__(self, root, session_id=0):
        OperatingSession.__init__(self, root, session_id)
        self.is_dummy = True
        self.canvas_saver.is_dummy = True

    def dummy_load_and_template(self, session_full_name, template_name):
        self.steps_order = [(self.preload, session_full_name),
                            self.take_place,
                            self.load,
                            (self.save_session_template, template_name, True)]
        self.next_function()

    def dummy_duplicate(self, path, args, src_addr):
        self.remember_osc_args(path, args, src_addr)
        session_to_load, new_session_full_name, sess_root = args
        self.steps_order = [(self.preload, session_to_load),
                            self.take_place,
                            self.load,
                            (self.duplicate, new_session_full_name),
                            self.duplicate_only_done]
        self.next_function()

    def ray_server_save_session_template(self, path, args, src_addr):
        self.remember_osc_args(path, args, src_addr)
        session_name, template_name, net = args
        self.steps_order = [(self.preload, session_name),
                            self.take_place,
                            self.load,
                            (self.save_session_template, template_name, net)]
        self.next_function()

    def ray_server_rename_session(self, path, args, src_addr):
        self.remember_osc_args(path, args, src_addr)
        full_session_name, new_session_name = args

        self.steps_order = [(self.preload, full_session_name),
                            self.take_place,
                            self.load,
                            (self.rename, new_session_name),
                            self.save,
                            (self.rename_done, new_session_name)]
        self.next_function()
    
    def ray_server_get_session_preview(self, path, args, src_addr,
                                       folder_sizes:list):
        session_name = args[0]
        self.steps_order = [(self.preload, session_name, False),
                            self.take_place,
                            self.load,
                            (self.send_preview, src_addr, folder_sizes)]
        self.next_function()
    
    def dummy_load(self, session_name):
        self.steps_order = [(self.preload, session_name, False),
                            self.take_place,
                            self.load]
        self.next_function()




try:
    import liblo
    from liblo import Message
except ImportError:
    import pyliblo3 as liblo
    from pyliblo3 import Message

def make_method(*args, **kwargs):
    return liblo.make_method(*args, **kwargs)


class Server(liblo.Server):
    url: str
    
    def __init__(self, *args, **kwargs):
        liblo.Server.__init__(self, *args, **kwargs)

    def add_method(self, *args, **kwargs):
        return liblo.Server.add_method(self, *args, **kwargs)
        
    def recv(self, *args, **kwargs):
        return liblo.Server.recv(self, *args, **kwargs)

    def send(self, *args, **kwargs):
        return liblo.Server.send(self, *args, **kwargs)


class ServerThread(liblo.ServerThread):
    url: str

    def __init__(self, *args, **kwargs):
        liblo.ServerThread.__init__(self, *args, **kwargs)

    def add_method(self, *args, **kwargs):
        return liblo.ServerThread.add_method(self, *args, **kwargs)
    
    def send(self, *args, **kwargs):
        return liblo.ServerThread.send(self, *args, **kwargs)

''' seems impossible to overinstance Message,
    it may be checked with type() function'''
class Message(liblo.Message):
    def __init__(self, *args, **kwargs):
        liblo.Message(self, *args, **kwargs)


class Address(liblo.Address):
    hostname: str
    url: str
    port: int
    
    def __init__(self, *args):
        liblo.Address.__init__(self, *args)
    



from .patchbay_manager import PatchbayManager, patchcanvas
from .base_elements import Group, Port, Portgroup
from .calbacker import Callbacker
from .tools_widgets import  PatchbayToolsWidget
from .port_info_dialog import CanvasPortInfoDialog
from .canvas_menu import CanvasMenu
from .options_dialog import CanvasOptionsDialog
from .filter_frame import FilterFrame
from .patchcanvas import patchcanvas
from .patchcanvas.scene_view import PatchGraphicsView


#!/usr/bin/python3 -u

import argparse
import os
import sys
import time
import signal
import subprocess
from PyQt5.QtCore import (QObject, QTimer, QProcess, QSettings,
                          QLocale, QTranslator, QFile)
from PyQt5.QtWidgets import (QApplication, QDialog, QFileDialog,
                             QMainWindow)
from PyQt5.QtXml import QDomDocument

import ray
from nsm_client_qt import NSMThread, NSMSignaler
import ui_proxy_gui
import ui_proxy_copy

ERR_OK = 0
ERR_NO_PROXY_FILE = -1
ERR_NOT_ENOUGHT_LINES = -2
ERR_NO_EXECUTABLE = -3
ERR_WRONG_ARGUMENTS = -4
ERR_WRONG_SAVE_SIGNAL = -5
ERR_WRONG_STOP_SIGNAL = -6



def signalHandler(sig, frame):
    if sig in (signal.SIGINT, signal.SIGTERM):
        if proxy.isRunning():
            proxy.waitForStop()
            proxy.stopProcess()
        else:
            #sys.exit()
            app.quit()



def ifDebug(string):
    if debug:
        sys.stderr.write(string + '\n')


class ProxyCopyDialog(QDialog):
    def __init__(self):
        QDialog.__init__(self)
        self.ui = ui_proxy_copy.Ui_Dialog()
        self.ui.setupUi(self)

        self.rename_file = False
        self.ui.pushButtonCopyRename.clicked.connect(self.setRenameFile)

    def setRenameFile(self):
        self.rename_file = True
        self.accept()

    def setFile(self, path):
        self.ui.labelFileNotInFolder.setText(
            _translate(
                'Dialog', '%s is not in proxy directory') %
            ('<strong>' + os.path.basename(path) + '</strong>'))


class ProxyDialog(QMainWindow):
    def __init__(self, executable=''):
        QMainWindow.__init__(self)
        self.ui = ui_proxy_gui.Ui_MainWindow()
        self.ui.setupUi(self)

        self.server = server
        self.proxy = proxy

        self.config_file = ''
        self.args_edited = False
        self.fields_allow_start = False
        self.process_is_running = False


        self.ui.toolButtonBrowse.clicked.connect(self.browseFile)

        self.ui.lineEditExecutable.textEdited.connect(
            self.lineEditExecutableEdited)
        self.ui.lineEditArguments.textChanged.connect(
            self.lineEditArgumentsChanged)
        self.ui.lineEditConfigFile.textChanged.connect(
            self.lineEditConfigFileChanged)

        self.ui.comboSaveSig.addItem(_translate('proxy', 'None'), 0)
        self.ui.comboSaveSig.addItem('SIGUSR1', int(signal.SIGUSR1))
        self.ui.comboSaveSig.addItem('SIGUSR2', int(signal.SIGUSR2))
        self.ui.comboSaveSig.addItem('SIGINT', int(signal.SIGINT))
        self.ui.comboSaveSig.activated.connect(self.comboSaveSigChanged)
        self.ui.comboSaveSig.setCurrentIndex(0)

        self.ui.comboNoSave.addItem(
            _translate('proxy', "0 - Ignore missing save"))
        self.ui.comboNoSave.addItem(
            _translate('proxy', "1 - Transmit missing save"))
        self.ui.comboNoSave.addItem(
            _translate('proxy', "2 - Accept windows close"))
        self.ui.comboNoSave.setCurrentIndex(2)
        self.ui.comboNoSave.activated.connect(self.comboNoSaveChanged)
        self.ui.labelNoSaveLevel.setToolTip(self.ui.comboNoSave.toolTip())

        self.ui.comboStopSig.addItem('SIGTERM', int(signal.SIGTERM))
        self.ui.comboStopSig.addItem('SIGINT', int(signal.SIGINT))
        self.ui.comboStopSig.addItem('SIGHUP', int(signal.SIGHUP))
        self.ui.comboStopSig.activated.connect(self.comboStopSigChanged)
        self.ui.comboStopSig.setCurrentIndex(0)

        self.ui.comboSaveSig.currentTextChanged.connect(self.allowSaveTest)
        self.ui.toolButtonTestSave.clicked.connect(self.testSave)
        self.ui.toolButtonTestSave.setEnabled(False)

        self.ui.pushButtonStart.clicked.connect(self.startProcess)
        self.ui.pushButtonStop.clicked.connect(self.stopProcess)
        self.ui.pushButtonStop.setEnabled(False)

        self.ui.lineEditExecutable.setText(executable)
        self.lineEditExecutableEdited(executable)

        self.ui.labelError.setText('')

        proxy.process.started.connect(self.proxyStarted)
        proxy.process.finished.connect(self.proxyFinished)
        if ray.QT_VERSION >= (5, 6):
            proxy.process.errorOccurred.connect(self.proxyErrorInProcess)

    def checkAllowStart(self):
        self.fields_allow_start = True
        if not self.ui.lineEditExecutable.text():
            self.fields_allow_start = False

        if ray.shell_line_to_args(self.ui.lineEditArguments.text()) is None:
            self.fields_allow_start = False

        self.ui.pushButtonStart.setEnabled(
            bool(not self.process_is_running and self.fields_allow_start))

    def updateValuesFromProxyFile(self):
        self.ui.lineEditExecutable.setText(proxy.executable)
        self.ui.lineEditConfigFile.setText(proxy.config_file)
        self.ui.lineEditArguments.setText(proxy.arguments_line)

        save_index = self.ui.comboSaveSig.findData(proxy.save_signal)
        self.ui.comboSaveSig.setCurrentIndex(save_index)

        self.ui.comboNoSave.setCurrentIndex(proxy.no_save_level)
        self.ui.comboNoSave.setEnabled(not bool(proxy.save_signal))

        stop_index = self.ui.comboStopSig.findData(proxy.stop_signal)
        self.ui.comboStopSig.setCurrentIndex(stop_index)

        self.ui.checkBoxWaitWindow.setChecked(proxy.wait_window)

        self.checkAllowStart()

    def browseFile(self):
        config_file, ok = QFileDialog.getOpenFileName(
            self, _translate('Dialog', 'Select File to use as CONFIG_FILE'))
        if not ok:
            return

        if not config_file.startswith(os.getcwd() + '/'):
            qfile = QFile(config_file)
            if qfile.size() < 20971520:  # if file < 20Mb
                copy_dialog = ProxyCopyDialog()
                copy_dialog.setFile(config_file)
                copy_dialog.exec()

                if copy_dialog.result():
                    if copy_dialog.rename_file:
                        base, pt, extension = os.path.basename(
                            config_file).rpartition('.')

                        config_file = "%s.%s" % (proxy.session_name, extension)
                        if not base:
                            config_file = proxy.session_name
                    else:
                        config_file = os.path.basename(config_file)

                    qfile.copy(config_file)

        self.config_file = os.path.relpath(config_file)
        if (proxy.session_name
            and (self.config_file == proxy.session_name
                 or self.config_file.startswith("%s." % proxy.session_name))):
            self.config_file = self.config_file.replace(proxy.session_name,
                                                        "$RAY_SESSION_NAME")
        self.ui.lineEditConfigFile.setText(self.config_file)

    def lineEditExecutableEdited(self, text):
        self.checkAllowStart()

    def lineEditArgumentsChanged(self, text):
        self.checkAllowStart()
        if ray.shell_line_to_args(text) is not None:
            self.ui.lineEditArguments.setStyleSheet('')
        else:
            self.ui.lineEditArguments.setStyleSheet(
                'QLineEdit{background: red}')
            self.ui.pushButtonStart.setEnabled(False)

    def lineEditConfigFileChanged(self, text):
        if text and not self.ui.lineEditArguments.text():
            self.ui.lineEditArguments.setText('"$CONFIG_FILE"')
        elif (not text
              and self.ui.lineEditArguments.text() == '"$CONFIG_FILE"'):
            self.ui.lineEditArguments.setText('')

    def comboSaveSigChanged(self, index):
        self.ui.comboNoSave.setEnabled(bool(index == 0))

        save_signal = 0

        if index == 1:
            save_signal = signal.SIGUSR1
        elif index == 2:
            save_signal = signal.SIGUSR2
        elif index == 3:
            save_signal = signal.SIGINT

        save_signal = int(save_signal)
        self.proxy.setSaveSignal(save_signal)

    def comboStopSigChanged(self, index):
        stop_signal = signal.SIGTERM

        if index == 1:
            stop_signal = signal.SIGINT
        elif index == 2:
            stop_signal = signal.SIGTERM

        stop_signal = int(stop_signal)
        self.proxy.setStopSignal(stop_signal)

    def comboNoSaveChanged(self, index):
        self.proxy.no_save_level = index
        self.proxy.sendNoSaveLevel()

    def allowSaveTest(self, text=None):
        if text is None:
            text = self.ui.comboSaveSig.currentText()

        self.ui.toolButtonTestSave.setEnabled(
            bool(self.process_is_running and text != 'None'))

    def testSave(self):
        save_signal = self.ui.comboSaveSig.currentData()
        proxy.saveProcess(save_signal)

    def saveProxy(self):
        executable = self.ui.lineEditExecutable.text()
        config_file = self.ui.lineEditConfigFile.text()
        arguments_line = self.ui.lineEditArguments.text()
        save_signal = self.ui.comboSaveSig.currentData()
        no_save_level = self.ui.comboNoSave.currentIndex()
        stop_signal = self.ui.comboStopSig.currentData()
        wait_window = self.ui.checkBoxWaitWindow.isChecked()

        proxy.updateAndSave(
            executable,
            config_file,
            arguments_line,
            save_signal,
            no_save_level,
            stop_signal,
            wait_window)

    def startProcess(self):
        self.saveProxy()

        if proxy.is_launchable:
            proxy.startProcess()

    def stopProcess(self):
        proxy.stopProcess(self.ui.comboStopSig.currentData())

    def proxyStarted(self):
        self.process_is_running = True
        self.ui.pushButtonStart.setEnabled(False)
        self.ui.pushButtonStop.setEnabled(True)
        self.allowSaveTest()
        self.ui.labelError.setText('')

    def processTerminateShortly(self, duration):
        self.ui.labelError.setText('Process terminate in %f ms')

    def proxyFinished(self):
        self.process_is_running = False
        self.ui.pushButtonStart.setEnabled(self.fields_allow_start)
        self.ui.pushButtonStop.setEnabled(False)
        self.allowSaveTest()
        self.ui.labelError.setText('')

    def proxyErrorInProcess(self):
        self.ui.labelError.setText(
            _translate(
                'Dialog',
                'Executable failed to launch ! It\'s maybe not present on system.'))
        if not self.isVisible():
            self.show()

    def closeEvent(self, event):
        server.sendToDaemon('/nsm/client/gui_is_hidden')
        settings.setValue(
            'ProxyGui%s/geometry' %
            self.proxy.client_id,
            self.saveGeometry())
        settings.setValue(
            'ProxyGui%s/WindowState' %
            self.proxy.client_id, self.saveState())
        settings.sync()

        if self.fields_allow_start:
            self.saveProxy()

        QMainWindow.closeEvent(self, event)

        # Quit if process is not running yet
        if not proxy.process.state() == QProcess.Running:
            sys.exit(0)

    def showEvent(self, event):
        self.server.sendToDaemon('/nsm/client/gui_is_shown')

        if settings.value('ProxyGui%s/geometry' % self.proxy.client_id):
            self.restoreGeometry(
                settings.value(
                    'ProxyGui%s/geometry' %
                    self.proxy.client_id))
        if settings.value('ProxyGui%s/WindowState' % self.proxy.client_id):
            self.restoreState(
                settings.value(
                    'ProxyGui%s/WindowState' %
                    self.proxy.client_id))

        self.updateValuesFromProxyFile()

        QMainWindow.showEvent(self, event)
##########################


class Proxy(QObject):
    def __init__(self, executable=''):
        QObject.__init__(self)
        self.process = QProcess()
        self.process.setProcessChannelMode(QProcess.ForwardedChannels)
        self.process.finished.connect(self.processFinished)

        #self.proxy_file = None
        self.is_launchable = False
        self.project_path = ""
        self.path = ""
        self.session_name = ""
        self.client_id = ""
        self.jack_client_name = ""

        self.executable = executable
        self.arguments = []
        self.arguments_line = ''
        self.config_file = ""
        self.save_signal = 0
        self.no_save_level = 2
        self.stop_signal = int(signal.SIGTERM)
        self.label = ""

        self.wait_window = False

        self._config_file_used = False
        self._wait_for_stop = False

        self.timer_save = QTimer()
        self.timer_save.setSingleShot(True)
        self.timer_save.setInterval(300)
        self.timer_save.timeout.connect(self.timerSaveFinished)

        self.timer_open = QTimer()
        self.timer_open.setSingleShot(True)
        self.timer_open.setInterval(500)
        self.timer_open.timeout.connect(self.timerOpenFinished)

        self.is_finishable = False
        self.timer_close = QTimer()
        self.timer_close.setSingleShot(True)
        self.timer_close.setInterval(2500)
        self.timer_close.timeout.connect(self.timerCloseFinished)
        self.timer_close.start()
        self.process_start_time = time.time()

        self.timer_window = QTimer()
        self.timer_window.setInterval(100)
        self.timer_window.timeout.connect(self.checkWindow)
        self.timer_window_n = 0

        signaler.server_sends_open.connect(self.initialize)
        signaler.server_sends_save.connect(self.saveProcess)
        signaler.show_optional_gui.connect(self.showOptionalGui)
        signaler.hide_optional_gui.connect(self.hideOptionalGui)

    def isRunning(self):
        return bool(self.process.state() == QProcess.Running)

    def waitForStop(self):
        self._wait_for_stop = True

    def setSaveSignal(self, int_signal):
        self.save_signal = int_signal
        self.sendNoSaveLevel()

    def setStopSignal(self, int_signal):
        self.stop_signal = int_signal

    def readFile(self):
        self.is_launchable = False
        try:
            file = open(self.path, 'r')
        except BaseException:
            return

        xml = QDomDocument()
        xml.setContent(file.read())

        content = xml.documentElement()

        if content.tagName() != "RAY-PROXY":
            file.close()
            return

        cte = content.toElement()
        file_version = cte.attribute('VERSION')
        self.executable = cte.attribute('executable')
        self.config_file = cte.attribute('config_file')
        self.arguments_line = cte.attribute('arguments')
        save_signal = cte.attribute('save_signal')
        no_save_level = cte.attribute('no_save_level')
        stop_signal = cte.attribute('stop_signal')

        wait_window = cte.attribute('wait_window')

        if wait_window.isdigit():
            self.wait_window = bool(int(wait_window))
        else:
            self.wait_window = False

        file.close()

        if save_signal.isdigit():
            self.save_signal = int(save_signal)

        if no_save_level.isdigit():
            self.no_save_level = int(no_save_level)
        else:
            self.no_save_level = 2

        versions = [file_version, '0.7.1']
        versions.sort()

        if file_version != versions[0]:
            # something was wrong in old version,
            # save signal was saved as stop signal too.
            # so don't read stop signal if this is an old file.
            if stop_signal.isdigit():
                self.stop_signal = int(stop_signal)

        if not self.executable:
            return

        arguments = ray.shell_line_to_args(self.arguments_line)
        if arguments is None:
            return

        self.is_launchable = True

    def saveFile(
            self,
            executable,
            config_file,
            arguments_line,
            save_signal,
            no_save_level,
            stop_signal,
            wait_window):
        try:
            file = open(self.path, 'w')
        except BaseException:
            return

        if not save_signal:
            save_signal = 0

        xml = QDomDocument()
        p = xml.createElement('RAY-PROXY')
        p.setAttribute('VERSION', ray.VERSION)
        p.setAttribute('executable', executable)
        p.setAttribute('arguments', arguments_line)
        p.setAttribute('config_file', config_file)
        p.setAttribute('save_signal', str(int(save_signal)))
        p.setAttribute('no_save_level', str(no_save_level))
        p.setAttribute('stop_signal', str(int(stop_signal)))
        p.setAttribute('wait_window', wait_window)

        xml.appendChild(p)

        contents = "<?xml version='1.0' encoding='UTF-8'?>\n"
        contents += "<!DOCTYPE RAY-PROXY>\n"
        contents += xml.toString()

        file.write(contents)
        file.close()

        self.readFile()

    def updateValues(self, executable, config_file, arguments_line,
                     save_signal, no_save_level, stop_signal, wait_window):
        self.executable = executable
        self.config_file = config_file
        self.arguments_line = arguments_line
        self.save_signal = save_signal
        self.no_save_level = no_save_level
        self.stop_signal = stop_signal
        self.wait_window = wait_window

    def saveProxyFile(self):
        self.saveFile(
            self.executable,
            self.config_file,
            self.arguments_line,
            self.save_signal,
            self.no_save_level,
            self.stop_signal,
            self.wait_window)

    def updateAndSave(self, executable, config_file, arguments_line,
                      save_signal, no_save_level, stop_signal, wait_window):
        self.updateValues(executable, config_file, arguments_line,
                          save_signal, no_save_level,
                          stop_signal, wait_window)
        self.saveProxyFile()

    def processFinished(self, exit_code):
        if self._wait_for_stop:
            app.quit()

        if self.is_finishable:
            if not proxy_dialog.isVisible():
                app.quit()
        else:
            duration = time.time() - self.process_start_time
            proxy_dialog.processTerminateShortly(duration)
            # proxy_dialog.show()

    def checkWindow(self):
        self.timer_window_n += 1

        if self.timer_window_n > 600:
            # 600 x 50ms = 30s max until ray-proxy
            # replyOpen to Session Manager
            self.checkWindowEnded()
            return

        try:
            # get all windows and their PID with wmctrl
            wmctrl_all = subprocess.check_output(
                ['wmctrl', '-l', '-p']).decode()
        except BaseException:
            self.checkWindowEnded()
            return

        if not wmctrl_all:
            self.checkWindowEnded()
            return

        all_lines = wmctrl_all.split('\n')
        pids = []

        # get all windows pids
        for line in all_lines:
            if not line:
                continue

            line_sep = line.split(' ')
            non_empt = []
            for el in line_sep:
                if el:
                    non_empt.append(el)

            if len(non_empt) >= 3 and non_empt[2].isdigit():
                pids.append(int(non_empt[2]))
            else:
                # window manager seems to not work correctly with wmctrl, so
                # replyOpen now
                self.checkWindowEnded()
                return

        parent_pid = self.process.pid()

        # check in pids if one comes from this ray-proxy
        for pid in pids:
            if pid < parent_pid:
                continue

            ppid = pid

            while ppid != parent_pid and ppid > 1:
                try:
                    proc_file = open('/proc/%i/status' % ppid, 'r')
                    proc_contents = proc_file.read()
                except BaseException:
                    self.checkWindowEnded()
                    return

                for line in proc_contents.split('\n'):
                    if line.startswith('PPid:'):
                        ppid_str = line.rpartition('\t')[2]
                        if ppid_str.isdigit():
                            ppid = int(ppid_str)
                            break
                else:
                    self.checkWindowEnded()
                    return

            if ppid == parent_pid:
                # a window appears with a pid child of this ray-proxy,
                # replyOpen
                QTimer.singleShot(200, self.checkWindowEnded)
                break

    def checkWindowEnded(self):
        self.timer_window.stop()
        server.openReply()

    def initialize(self, project_path, session_name, jack_client_name):
        self.project_path = project_path
        self.session_name = session_name
        self.client_id = project_path.rpartition('.')[2]
        self.jack_client_name = jack_client_name

        server.sendGuiState(False)

        if not os.path.exists(project_path):
            os.mkdir(project_path)

        os.chdir(project_path)

        proxy_dialog.setWindowTitle("Ray Proxy - %s" % self.client_id)

        self.path = os.path.join(project_path, "ray-proxy.xml")
        self.readFile()

        proxy_dialog.updateValuesFromProxyFile()

        if not self.is_launchable:
            server.openReply()
            proxy_dialog.show()
            return

        self.startProcess()

    def sendNoSaveLevel(self):
        if ':no-save-level:' in server.getServerCapabilities():
            nsl = self.no_save_level
            if self.save_signal:
                nsl = 0
            server.sendToDaemon('/nsm/client/no_save_level', nsl)

    def startProcess(self):
        os.environ['NSM_CLIENT_ID'] = self.jack_client_name
        os.environ['RAY_CLIENT_ID'] = self.client_id
        os.environ['RAY_SESSION_NAME'] = self.session_name

        # enable environment vars in config_file
        config_file = os.path.expandvars(self.config_file)
        os.environ['CONFIG_FILE'] = config_file

        # because that is not done by python itself
        os.environ['PWD'] = os.getcwd()

        # Useful for launching NSM compatible clients with specifics arguments
        nsm_url = os.getenv('NSM_URL')
        ray_port_str = nsm_url.rpartition(':')[2]
        if ray_port_str.endswith('/'):
            ray_port_str = ray_port_str[:-1]
        if ray_port_str.isdigit():
            os.environ['RAY_CONTROL_PORT'] = ray_port_str

        os.unsetenv('NSM_URL')

        arguments_line = os.path.expandvars(self.arguments_line)
        arguments = ray.shell_line_to_args(arguments_line)

        self._config_file_used = bool(config_file
                                      and config_file in arguments)
        self.sendNoSaveLevel()

        self.process.start(self.executable, arguments)
        self.timer_open.start()

    def saveProcess(self, save_signal=0):
        if not save_signal:
            save_signal = self.save_signal

        if self.isRunning() and save_signal:
            os.kill(self.process.processId(), save_signal)

        self.timer_save.start()

    def stopProcess(self, signal=signal.SIGTERM):
        if signal is None:
            return

        if not self.isRunning():
            return

        os.kill(self.process.processId(), signal)

    def timerSaveFinished(self):
        server.saveReply()

    def timerOpenFinished(self):
        if self.wait_window:
            self.timer_window.start()
        else:
            server.openReply()

        if self.isRunning() and proxy_dialog.isVisible():
            proxy_dialog.close()

    def timerCloseFinished(self):
        self.is_finishable = True

    def stop(self):
        if self.process.state:
            self.process.terminate()

    def showOptionalGui(self):
        proxy_dialog.show()

    def hideOptionalGui(self):
        if not proxy_dialog.isHidden():
            proxy_dialog.close()


class ProxyFile:
    def __init__(self, project_path, executable=''):


        self.executable = executable
        self.arguments_line = ''
        self.config_file = ''
        self.args_line = ''
        self.save_signal = 0
        self.stop_signal = int(signal.SIGTERM)
        self.wait_window = False


if __name__ == '__main__':
    NSM_URL = os.getenv('NSM_URL')
    if not NSM_URL:
        sys.stderr.write('Could not register as NSM client.\n')
        sys.exit()

    daemon_address = ray.get_liblo_address(NSM_URL)

    parser = argparse.ArgumentParser()
    parser.add_argument('--executable', default='')
    parser.add_argument('--debug',
                        '-d',
                        action='store_true',
                        help='see all OSC messages')
    parser.add_argument('-v', '--version', action='version',
                        version=ray.VERSION)
    parsed_args = parser.parse_args()

    debug = parsed_args.debug
    executable = parsed_args.executable

    app = QApplication(sys.argv)
    app.setApplicationName("RaySession")
    # app.setApplicationVersion(ray.VERSION)
    app.setOrganizationName("RaySession")
    app.setQuitOnLastWindowClosed(False)
    settings = QSettings()

    signal.signal(signal.SIGINT, signalHandler)
    signal.signal(signal.SIGTERM, signalHandler)

    # Translation process
    locale = QLocale.system().name()
    appTranslator = QTranslator()

    if appTranslator.load(
        "%s/locale/raysession_%s" %
        (os.path.dirname(
            os.path.dirname(
                os.path.dirname(
                    sys.argv[0]))),
         locale)):
        app.installTranslator(appTranslator)
    _translate = app.translate

    timer = QTimer()
    timer.setInterval(200)
    timer.timeout.connect(lambda: None)
    timer.start()

    signaler = NSMSignaler()

    server = NSMThread('ray-proxy', signaler,
                       daemon_address, debug)
    server.start()

    proxy = Proxy(executable)
    proxy_dialog = ProxyDialog()


    server.announce('Ray Proxy', ':optional-gui:warning-no-save:', 'ray-proxy')

    app.exec()

    settings.sync()
    server.stop()

    del server
    del proxy
    del app
